import numpy as np

from relsad.network.components import (
    Battery,
    Bus,
    CircuitBreaker,
    Disconnector,
    EVPark,
    IntelligentSwitch,
    Line,
    MainController,
    ManualMainController,
    MicrogridMode,
    Production,
    Sensor,
)
from relsad.network.systems import (
    Distribution,
    Microgrid,
    PowerSystem,
    Transmission,
)
from relsad.StatDist import NormalParameters, StatDist, StatDistType
from relsad.Table import Table
from relsad.Time import Time, TimeUnit
from relsad.visualization.plotting import plot_topology


def initialize_network(
    include_microgrid: bool = True,
    include_production: bool = True,
    include_backup: bool = True,
    include_ICT: bool = True,
    include_ev: bool = True,
    v2g_flag: bool = True,
    fail_rate_trafo: float = 0.0,  # fails per year
    fail_rate_line: float = 0.013,  # fails per year
    inverter: float = 0.0036,
    ev_percentage: float = 0.46,
    ev_E_max: float = 0.07,
    line_repair_time_stat_dist: StatDist = StatDist(
        stat_dist_type=StatDistType.TRUNCNORMAL,
        parameters=NormalParameters(
            loc=1.25,
            scale=1,
            min_val=0.5,
            max_val=2,
        ),
    ),
    microgrid_mode: MicrogridMode = MicrogridMode.LIMITED_SUPPORT,
    battery_capacity: float = 1,  # MWh
):
    def num_ev_table_func(
        n_customers,
        ev_percentage=0.47,
        daily_charge_frac=0.61,
    ):
        return Table(
            x=np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    9,
                    10,
                    11,
                    12,
                    13,
                    14,
                    15,
                    16,
                    17,
                    18,
                    19,
                    20,
                    21,
                    22,
                    23,
                ],
            ),
            y=np.array(
                [
                    0.52,
                    0.52,
                    0.52,
                    0.52,
                    0.52,
                    0.52,
                    0.52,
                    0.08,
                    0.08,
                    0.18,
                    0.18,
                    0.18,
                    0.18,
                    0.18,
                    0.18,
                    0.18,
                    0.28,
                    0.28,
                    0.28,
                    0.28,
                    0.42,
                    0.42,
                    0.42,
                    0.42,
                ]
            )
            * n_customers
            * ev_percentage
            * daily_charge_frac,
        )

    if include_ICT:
        C1 = MainController(
            name="C1",
        )
    else:
        C1 = ManualMainController(
            name="C1",
            sectioning_time=Time(1, TimeUnit.HOUR),
        )

    ps = PowerSystem(C1)

    B1 = Bus("B1", n_customers=0, coordinate=[-1, 0], fail_rate_per_year=0)
    B2 = Bus(
        "B2",
        n_customers=1,
        coordinate=[1, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B3 = Bus(
        "B3",
        n_customers=39,
        coordinate=[2, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B4 = Bus(
        "B4",
        n_customers=1,
        coordinate=[3, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B5 = Bus(
        "B5",
        n_customers=26,
        coordinate=[4, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B6 = Bus(
        "B6",
        n_customers=26,
        coordinate=[5, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B7 = Bus(
        "B7",
        n_customers=1,
        coordinate=[6, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B8 = Bus(
        "B8",
        n_customers=1,
        coordinate=[7, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B9 = Bus(
        "B9",
        n_customers=26,
        coordinate=[8, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B10 = Bus(
        "B10",
        n_customers=26,
        coordinate=[9, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B11 = Bus(
        "B11",
        n_customers=19,
        coordinate=[10, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B12 = Bus(
        "B12",
        n_customers=26,
        coordinate=[11, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B13 = Bus(
        "B13",
        n_customers=26,
        coordinate=[12, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B14 = Bus(
        "B14",
        n_customers=1,
        coordinate=[13, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B15 = Bus(
        "B15",
        n_customers=26,
        coordinate=[14, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B16 = Bus(
        "B16",
        n_customers=26,
        coordinate=[15, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B17 = Bus(
        "B17",
        n_customers=26,
        coordinate=[16, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B18 = Bus(
        "B18",
        n_customers=39,
        coordinate=[17, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B19 = Bus(
        "B19",
        n_customers=39,
        coordinate=[2, 1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B20 = Bus(
        "B20",
        n_customers=39,
        coordinate=[3, 1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B21 = Bus(
        "B21",
        n_customers=39,
        coordinate=[4, 1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B22 = Bus(
        "B22",
        n_customers=39,
        coordinate=[5, 1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B23 = Bus(
        "B23",
        n_customers=39,
        coordinate=[3, -1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B24 = Bus(
        "B24",
        n_customers=2,
        coordinate=[4, -1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B25 = Bus(
        "B25",
        n_customers=2,
        coordinate=[5, -1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B26 = Bus(
        "B26",
        n_customers=26,
        coordinate=[6, 1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B27 = Bus(
        "B27",
        n_customers=26,
        coordinate=[7, 1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B28 = Bus(
        "B28",
        n_customers=26,
        coordinate=[8, 1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B29 = Bus(
        "B29",
        n_customers=1,
        coordinate=[9, 1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B30 = Bus(
        "B30",
        n_customers=1,
        coordinate=[10, 1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B31 = Bus(
        "B31",
        n_customers=1,
        coordinate=[11, 1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B32 = Bus(
        "B32",
        n_customers=1,
        coordinate=[12, 1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B33 = Bus(
        "B33",
        n_customers=26,
        coordinate=[13, 1],
        fail_rate_per_year=fail_rate_trafo,
    )

    L1 = Line(
        "L1",
        B1,
        B2,
        r=0.0922,
        x=0.0470,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L2 = Line(
        "L2",
        B2,
        B3,
        r=0.4930,
        x=0.2511,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L3 = Line(
        "L3",
        B3,
        B4,
        r=0.3660,
        x=0.1864,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L4 = Line(
        "L4",
        B4,
        B5,
        r=0.3811,
        x=0.1941,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L5 = Line(
        "L5",
        B5,
        B6,
        r=0.8190,
        x=0.7070,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L6 = Line(
        "L6",
        B6,
        B7,
        r=0.1872,
        x=0.6188,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L7 = Line(
        "L7",
        B7,
        B8,
        r=0.7114,
        x=0.2351,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L8 = Line(
        "L8",
        B8,
        B9,
        r=1.0300,
        x=0.7400,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L9 = Line(
        "L9",
        B9,
        B10,
        r=1.0440,
        x=0.7400,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L10 = Line(
        "L10",
        B10,
        B11,
        r=0.1966,
        x=0.0650,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L11 = Line(
        "L11",
        B11,
        B12,
        r=0.3744,
        x=0.1238,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L12 = Line(
        "L12",
        B12,
        B13,
        r=1.4680,
        x=1.1550,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L13 = Line(
        "L13",
        B13,
        B14,
        r=0.5416,
        x=0.7129,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L14 = Line(
        "L14",
        B14,
        B15,
        r=0.5910,
        x=0.5260,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L15 = Line(
        "L15",
        B15,
        B16,
        r=0.7463,
        x=0.5450,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L16 = Line(
        "L16",
        B16,
        B17,
        r=1.2890,
        x=1.72010,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L17 = Line(
        "L17",
        B17,
        B18,
        r=0.7320,
        x=0.5740,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L18 = Line(
        "L18",
        B2,
        B19,
        r=0.1640,
        x=0.1565,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L19 = Line(
        "L19",
        B19,
        B20,
        r=1.5042,
        x=1.3554,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L20 = Line(
        "L20",
        B20,
        B21,
        r=0.4095,
        x=0.4784,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L21 = Line(
        "L21",
        B21,
        B22,
        r=0.7089,
        x=0.9373,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L22 = Line(
        "L22",
        B3,
        B23,
        r=0.4512,
        x=0.3083,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L23 = Line(
        "L23",
        B23,
        B24,
        r=0.8980,
        x=0.7091,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L24 = Line(
        "L24",
        B24,
        B25,
        r=0.8960,
        x=0.7011,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L25 = Line(
        "L25",
        B6,
        B26,
        r=0.2030,
        x=0.1034,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L26 = Line(
        "L26",
        B26,
        B27,
        r=0.2842,
        x=0.1447,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L27 = Line(
        "L27",
        B27,
        B28,
        r=1.0590,
        x=0.9337,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L28 = Line(
        "L28",
        B28,
        B29,
        r=0.8042,
        x=0.7006,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L29 = Line(
        "L29",
        B29,
        B30,
        r=0.5075,
        x=0.2585,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L30 = Line(
        "L30",
        B30,
        B31,
        r=0.9744,
        x=0.9630,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L31 = Line(
        "L31",
        B31,
        B32,
        r=0.3105,
        x=0.3619,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L32 = Line(
        "L32",
        B32,
        B33,
        r=0.3410,
        x=0.5320,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )

    CircuitBreaker("E1", L1)

    DL1a = Disconnector("L1a", L1, B1)
    DL1b = Disconnector("L1b", L1, B2)
    DL2a = Disconnector("L2a", L2, B2)
    DL2b = Disconnector("L2b", L2, B3)
    DL3a = Disconnector("L3a", L3, B3)
    DL3b = Disconnector("L3b", L3, B4)
    DL4a = Disconnector("L4a", L4, B4)
    DL4b = Disconnector("L4b", L4, B5)
    DL5a = Disconnector("L5a", L5, B5)
    DL5b = Disconnector("L5b", L5, B6)
    DL6a = Disconnector("L6a", L6, B6)
    DL6b = Disconnector("L6b", L6, B7)
    DL7a = Disconnector("L7a", L7, B7)
    DL7b = Disconnector("L7b", L7, B8)
    DL8a = Disconnector("L8a", L8, B8)
    DL8b = Disconnector("L8b", L8, B9)
    DL9a = Disconnector("L9a", L9, B9)
    DL9b = Disconnector("L9b", L9, B10)
    DL10a = Disconnector("L10a", L10, B10)
    DL10b = Disconnector("L10b", L10, B11)
    DL11a = Disconnector("L11a", L11, B11)
    DL11b = Disconnector("L11b", L11, B12)
    DL12a = Disconnector("L12a", L12, B12)
    DL12b = Disconnector("L12b", L12, B13)
    DL13a = Disconnector("L13a", L13, B13)
    DL13b = Disconnector("L13b", L13, B14)
    DL14a = Disconnector("L14a", L14, B14)
    DL14b = Disconnector("L14b", L14, B15)
    DL15a = Disconnector("L15a", L15, B15)
    DL15b = Disconnector("L15b", L15, B16)
    DL16a = Disconnector("L16a", L16, B16)
    DL16b = Disconnector("L16b", L16, B17)
    DL17a = Disconnector("L17a", L17, B17)
    DL17b = Disconnector("L17b", L17, B18)

    DL18a = Disconnector("L18a", L18, B2)
    DL18b = Disconnector("L18b", L18, B19)
    DL19a = Disconnector("L19a", L19, B19)
    DL19b = Disconnector("L19b", L19, B20)
    DL20a = Disconnector("L20a", L20, B20)
    DL20b = Disconnector("L20b", L20, B21)
    DL21a = Disconnector("L21a", L21, B21)
    DL21b = Disconnector("L21b", L21, B22)

    DL22a = Disconnector("L22a", L22, B3)
    DL22b = Disconnector("L22b", L22, B23)
    DL23a = Disconnector("L23a", L23, B23)
    DL23b = Disconnector("L23b", L23, B24)
    DL24a = Disconnector("L24a", L24, B24)
    DL24b = Disconnector("L24b", L24, B25)

    DL25a = Disconnector("L25a", L25, B6)
    DL25b = Disconnector("L25b", L25, B26)
    DL26a = Disconnector("L26a", L26, B26)
    DL26b = Disconnector("L26b", L26, B27)
    DL27a = Disconnector("L27a", L27, B27)
    DL27b = Disconnector("L27b", L27, B28)
    DL28a = Disconnector("L28a", L28, B28)
    DL28b = Disconnector("L28b", L28, B29)
    DL29a = Disconnector("L29a", L29, B29)
    DL29b = Disconnector("L29b", L29, B30)
    DL30a = Disconnector("L30a", L30, B30)
    DL30b = Disconnector("L30b", L30, B31)
    DL31a = Disconnector("L31a", L31, B31)
    DL31b = Disconnector("L31b", L31, B32)
    DL32a = Disconnector("L32a", L32, B32)
    DL32b = Disconnector("L32b", L32, B33)

    # Backup lines
    if include_backup:
        L33 = Line(
            "L33",
            B9,
            B15,
            r=2.0000,
            x=2.0000,
            fail_rate_density_per_year=fail_rate_line,
        )
        L34 = Line(
            "L34",
            B12,
            B22,
            r=2.0000,
            x=2.0000,
            fail_rate_density_per_year=fail_rate_line,
        )
        L35 = Line(
            "L35",
            B18,
            B33,
            r=0.5000,
            x=0.5000,
            fail_rate_density_per_year=fail_rate_line,
        )
        L36 = Line(
            "L36",
            B25,
            B29,
            r=0.5000,
            x=0.5000,
            fail_rate_density_per_year=fail_rate_line,
        )
        L37 = Line(
            "L37",
            B8,
            B21,
            r=2.0000,
            x=2.0000,
            fail_rate_density_per_year=fail_rate_line,
        )

        # Backup
        DL33a = Disconnector("L33a", L33, B9)
        DL33b = Disconnector("L33b", L33, B15)
        DL34a = Disconnector("L34a", L34, B12)
        DL34b = Disconnector("L34b", L34, B22)
        DL35a = Disconnector("L35a", L35, B18)
        DL35b = Disconnector("L35b", L35, B33)
        DL36a = Disconnector("L36a", L36, B25)
        DL36b = Disconnector("L36b", L36, B29)
        DL37a = Disconnector("L37a", L37, B8)
        DL37b = Disconnector("L37b", L37, B21)

        L33.set_backup()
        L34.set_backup()
        L35.set_backup()
        L36.set_backup()
        L37.set_backup()

    if include_production:

        Battery("Battery", B30)

    if include_ev:

        EVPark(
            name="EV1",
            bus=B3,
            num_ev_dist=num_ev_table_func(B3.n_customers),
            inj_p_max=inverter,
            E_max=ev_E_max,
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EV2",
            bus=B5,
            num_ev_dist=num_ev_table_func(B5.n_customers),
            inj_p_max=inverter,
            E_max=ev_E_max,
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EV3",
            bus=B6,
            num_ev_dist=num_ev_table_func(B6.n_customers),
            inj_p_max=inverter,
            E_max=ev_E_max,
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EV4",
            bus=B9,
            num_ev_dist=num_ev_table_func(B9.n_customers),
            inj_p_max=inverter,
            E_max=ev_E_max,
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EV5",
            bus=B10,
            num_ev_dist=num_ev_table_func(B10.n_customers),
            inj_p_max=inverter,
            E_max=ev_E_max,
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EV6",
            bus=B11,
            num_ev_dist=num_ev_table_func(B11.n_customers),
            inj_p_max=inverter,
            E_max=ev_E_max,
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EV7",
            bus=B12,
            num_ev_dist=num_ev_table_func(B12.n_customers),
            inj_p_max=inverter,
            E_max=ev_E_max,
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EV8",
            bus=B13,
            num_ev_dist=num_ev_table_func(B13.n_customers),
            inj_p_max=inverter,
            E_max=ev_E_max,
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EV9",
            bus=B15,
            num_ev_dist=num_ev_table_func(B15.n_customers),
            inj_p_max=inverter,
            E_max=ev_E_max,
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EV10",
            bus=B16,
            num_ev_dist=num_ev_table_func(B16.n_customers),
            inj_p_max=inverter,
            E_max=ev_E_max,
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EV11",
            bus=B17,
            num_ev_dist=num_ev_table_func(B17.n_customers),
            inj_p_max=inverter,
            E_max=ev_E_max,
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EV12",
            bus=B18,
            num_ev_dist=num_ev_table_func(B18.n_customers),
            inj_p_max=inverter,
            E_max=ev_E_max,
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EV13",
            bus=B19,
            num_ev_dist=num_ev_table_func(B19.n_customers),
            inj_p_max=inverter,
            E_max=ev_E_max,
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EV14",
            bus=B20,
            num_ev_dist=num_ev_table_func(B20.n_customers),
            inj_p_max=inverter,
            E_max=ev_E_max,
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EV15",
            bus=B21,
            num_ev_dist=num_ev_table_func(B21.n_customers),
            inj_p_max=inverter,
            E_max=ev_E_max,
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EV16",
            bus=B22,
            num_ev_dist=num_ev_table_func(B22.n_customers),
            inj_p_max=inverter,
            E_max=ev_E_max,
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EV17",
            bus=B23,
            num_ev_dist=num_ev_table_func(B23.n_customers),
            inj_p_max=inverter,
            E_max=ev_E_max,
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EV18",
            bus=B26,
            num_ev_dist=num_ev_table_func(B26.n_customers),
            inj_p_max=inverter,
            E_max=ev_E_max,
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EV19",
            bus=B27,
            num_ev_dist=num_ev_table_func(B27.n_customers),
            inj_p_max=inverter,
            E_max=ev_E_max,
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EV20",
            bus=B28,
            num_ev_dist=num_ev_table_func(B28.n_customers),
            inj_p_max=inverter,
            E_max=ev_E_max,
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EV21",
            bus=B33,
            num_ev_dist=num_ev_table_func(B33.n_customers),
            inj_p_max=inverter,
            E_max=ev_E_max,
            v2g_flag=v2g_flag,
        )

    if include_microgrid:

        # Micorgrid:

        BM1 = Bus(
            "BM1",
            n_customers=0,
            coordinate=[14, 2],
            fail_rate_per_year=fail_rate_trafo,
        )
        BM2 = Bus(
            "BM2",
            n_customers=40,
            coordinate=[15, 2],
            fail_rate_per_year=fail_rate_trafo,
        )
        BM3 = Bus(
            "BM3",
            n_customers=0,
            coordinate=[15, 2.5],
            fail_rate_per_year=fail_rate_trafo,
        )
        BM4 = Bus(
            "BM4",
            n_customers=0,
            coordinate=[15, 1.5],
            fail_rate_per_year=fail_rate_trafo,
        )

        Battery("Bat1", BM1, E_max=battery_capacity)
        Production("P1", BM3)
        Production("P2", BM4)

        # Microgrid

        ML1 = Line(
            "ML1",
            B33,
            BM1,
            r=0.1872,  # 0.7394,
            x=0.0619,  # 0.2444,
            repair_time_dist=line_repair_time_stat_dist,
            fail_rate_density_per_year=fail_rate_line,
        )

        ML2 = Line(
            "ML2",
            BM1,
            BM2,
            r=0.0047,
            x=0.0016,
            repair_time_dist=line_repair_time_stat_dist,
            fail_rate_density_per_year=fail_rate_line,
        )
        ML3 = Line(
            "ML3",
            BM1,
            BM3,
            r=0.0047,
            x=0.0016,
            repair_time_dist=line_repair_time_stat_dist,
            fail_rate_density_per_year=fail_rate_line,
        )
        ML4 = Line(
            "ML4",
            BM1,
            BM4,
            r=0.0047,
            x=0.0016,
            repair_time_dist=line_repair_time_stat_dist,
            fail_rate_density_per_year=fail_rate_line,
        )

        CircuitBreaker("E2", ML1)

        DML1a = Disconnector("ML1a", ML1, B33)
        DML1b = Disconnector("ML1b", ML1, BM1)
        DML2a = Disconnector("ML2a", ML2, BM1)
        DML2b = Disconnector("ML2b", ML2, BM2)
        DML3a = Disconnector("ML3a", ML3, BM1)
        DML3b = Disconnector("ML3b", ML3, BM3)
        DML4a = Disconnector("ML4a", ML4, BM1)
        DML4b = Disconnector("ML4b", ML4, BM4)

    if include_ICT:

        Sensor("SL1", L1)
        Sensor("SL2", L2)
        Sensor("SL3", L3)
        Sensor("SL4", L4)
        Sensor("SL5", L5)
        Sensor("SL6", L6)
        Sensor("SL7", L7)
        Sensor("SL8", L8)
        Sensor("SL9", L9)
        Sensor("SL10", L10)
        Sensor("SL11", L11)
        Sensor("SL12", L12)
        Sensor("SL13", L13)
        Sensor("SL14", L14)
        Sensor("SL15", L15)
        Sensor("SL16", L16)
        Sensor("SL17", L17)
        Sensor("SL18", L18)
        Sensor("SL19", L19)
        Sensor("SL20", L20)
        Sensor("SL21", L21)
        Sensor("SL22", L22)
        Sensor("SL23", L23)
        Sensor("SL24", L24)
        Sensor("SL25", L25)
        Sensor("SL26", L26)
        Sensor("SL27", L27)
        Sensor("SL28", L28)
        Sensor("SL29", L29)
        Sensor("SL30", L30)
        Sensor("SL31", L31)
        Sensor("SL32", L32)

        # Backup lines
        if include_backup:
            Sensor("SL33", L33)
            Sensor("SL34", L34)
            Sensor("SL35", L35)
            Sensor("SL36", L36)
            Sensor("SL37", L37)

        IntelligentSwitch("ISwL1a", DL1a)
        IntelligentSwitch("ISwL1b", DL1b)
        IntelligentSwitch("ISwL2a", DL2a)
        IntelligentSwitch("ISwL2b", DL2b)
        IntelligentSwitch("ISwL3a", DL3a)
        IntelligentSwitch("ISwL3b", DL3b)
        IntelligentSwitch("ISwL4a", DL4a)
        IntelligentSwitch("ISwL4b", DL4b)
        IntelligentSwitch("ISwL5a", DL5a)
        IntelligentSwitch("ISwL5b", DL5b)
        IntelligentSwitch("ISwL6a", DL6a)
        IntelligentSwitch("ISwL6b", DL6b)
        IntelligentSwitch("ISwL7a", DL7a)
        IntelligentSwitch("ISwL7b", DL7b)
        IntelligentSwitch("ISwL8a", DL8a)
        IntelligentSwitch("ISwL8b", DL8b)
        IntelligentSwitch("ISwL9a", DL9a)
        IntelligentSwitch("ISwL9b", DL9b)
        IntelligentSwitch("ISwL10a", DL10a)
        IntelligentSwitch("ISwL10b", DL10b)
        IntelligentSwitch("ISwL11a", DL11a)
        IntelligentSwitch("ISwL11b", DL11b)
        IntelligentSwitch("ISwL12a", DL12a)
        IntelligentSwitch("ISwL12b", DL12b)
        IntelligentSwitch("ISwL13a", DL13a)
        IntelligentSwitch("ISwL13b", DL13b)
        IntelligentSwitch("ISwL14a", DL14a)
        IntelligentSwitch("ISwL14b", DL14b)
        IntelligentSwitch("ISwL15a", DL15a)
        IntelligentSwitch("ISwL15b", DL15b)
        IntelligentSwitch("ISwL16a", DL16a)
        IntelligentSwitch("ISwL16b", DL16b)
        IntelligentSwitch("ISwL17a", DL17a)
        IntelligentSwitch("ISwL17b", DL17b)
        IntelligentSwitch("ISwL18a", DL18a)
        IntelligentSwitch("ISwL18b", DL18b)
        IntelligentSwitch("ISwL19a", DL19a)
        IntelligentSwitch("ISwL19b", DL19b)
        IntelligentSwitch("ISwL20a", DL20a)
        IntelligentSwitch("ISwL20b", DL20b)
        IntelligentSwitch("ISwL21a", DL21a)
        IntelligentSwitch("ISwL21b", DL21b)
        IntelligentSwitch("ISwL22a", DL22a)
        IntelligentSwitch("ISwL22b", DL22b)
        IntelligentSwitch("ISwL23a", DL23a)
        IntelligentSwitch("ISwL23b", DL23b)
        IntelligentSwitch("ISwL24a", DL24a)
        IntelligentSwitch("ISwL24b", DL24b)
        IntelligentSwitch("ISwL25a", DL25a)
        IntelligentSwitch("ISwL25b", DL25b)
        IntelligentSwitch("ISwL26a", DL26a)
        IntelligentSwitch("ISwL26b", DL26b)
        IntelligentSwitch("ISwL27a", DL27a)
        IntelligentSwitch("ISwL27b", DL27b)
        IntelligentSwitch("ISwL28a", DL28a)
        IntelligentSwitch("ISwL28b", DL28b)
        IntelligentSwitch("ISwL29a", DL29a)
        IntelligentSwitch("ISwL29b", DL29b)
        IntelligentSwitch("ISwL30a", DL30a)
        IntelligentSwitch("ISwL30b", DL30b)
        IntelligentSwitch("ISwL31a", DL31a)
        IntelligentSwitch("ISwL31b", DL31b)
        IntelligentSwitch("ISwL32a", DL32a)
        IntelligentSwitch("ISwL32b", DL32b)

        # Backup lines:
        if include_backup:
            IntelligentSwitch("ISwL33a", DL33a)
            IntelligentSwitch("ISwL33b", DL33b)
            IntelligentSwitch("ISwL34a", DL34a)
            IntelligentSwitch("ISwL34b", DL34b)
            IntelligentSwitch("ISwL35a", DL35a)
            IntelligentSwitch("ISwL35b", DL35b)
            IntelligentSwitch("ISwL36a", DL36a)
            IntelligentSwitch("ISwL36b", DL36b)
            IntelligentSwitch("ISwL37a", DL37a)
            IntelligentSwitch("ISwL37b", DL37b)

        if include_microgrid:
            Sensor("SML1", ML1)
            Sensor("SML2", ML2)
            Sensor("SML3", ML3)
            Sensor("SML4", ML4)

            IntelligentSwitch("ISwML1a", DML1a)
            IntelligentSwitch("ISwML1b", DML1b)
            IntelligentSwitch("ISwML2a", DML2a)
            IntelligentSwitch("ISwML2b", DML2b)
            IntelligentSwitch("ISwML3a", DML3a)
            IntelligentSwitch("ISwML3b", DML3b)
            IntelligentSwitch("ISwML4a", DML4a)
            IntelligentSwitch("ISwML4b", DML4b)

    tn = Transmission(ps, B1)

    dn = Distribution(tn, L1)

    dn.add_buses(
        [
            B2,
            B3,
            B4,
            B5,
            B6,
            B7,
            B8,
            B9,
            B10,
            B11,
            B12,
            B13,
            B14,
            B15,
            B16,
            B17,
            B18,
            B19,
            B20,
            B21,
            B22,
            B23,
            B24,
            B25,
            B26,
            B27,
            B28,
            B29,
            B30,
            B31,
            B32,
            B33,
        ]
    )

    dn.add_lines(
        [
            L2,
            L3,
            L4,
            L5,
            L6,
            L7,
            L8,
            L9,
            L10,
            L11,
            L12,
            L13,
            L14,
            L15,
            L16,
            L17,
            L18,
            L19,
            L20,
            L21,
            L22,
            L23,
            L24,
            L25,
            L26,
            L27,
            L28,
            L29,
            L30,
            L31,
            L32,
        ]
    )

    if include_backup:
        dn.add_lines(
            [
                L33,
                L34,
                L35,
                L36,
                L37,
            ]
        )

    if include_microgrid:
        m = Microgrid(dn, ML1, mode=microgrid_mode)
        m.add_buses([BM1, BM2, BM3, BM4])
        m.add_lines([ML2, ML3, ML4])

    return ps


if __name__ == "__main__":
    import os

    ps, _, _ = initialize_network()
    fig = plot_topology(
        buses=ps.buses,
        lines=ps.lines,
        figsize=(8, 2),
    )

    fig.savefig(
        os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "IEEE33.pdf",
        )
    )



from .monte_carlo.history import (
    merge_monte_carlo_history,
    merge_monte_carlo_child_network_history,
    merge_monte_carlo_comp_history,
)

from .sequence.history import (
    plot_obj_history,
    save_obj_history,
    save_sequence_history,
)

from .Simulation import Simulation

from .system_config import (
    find_sub_systems,
    try_to_add_connected_lines,
    add_bus,
    update_backup_lines_between_sub_systems,
    update_sub_system_slack,
    set_slack,
    prepare_system,
    reset_system,
)

__all__ = []
for v in dir():
    if not v.startswith("__") and v != "relsad":
        __all__.append(v)



import os
import time

from relsad.examples.CINELDI.load_and_prod import set_network_load_and_prod
from relsad.examples.CINELDI.network import initialize_network
from relsad.load.bus import CostFunction
from relsad.simulation import Simulation
from relsad.StatDist import (
    GammaParameters,
    NormalParameters,
    StatDist,
    StatDistType,
)
from relsad.Time import Time, TimeStamp, TimeUnit


def run_simulation(
    include_microgrid=True,
    include_production=True,
    include_ev: bool = True,
    v2g_flag: bool = True,
    line_repair_time_stat_dist: StatDist = StatDist(
        stat_dist_type=StatDistType.TRUNCNORMAL,
        parameters=NormalParameters(
            loc=1,
            scale=1,
            min_val=0,
            max_val=2,
        ),
    ),
    fail_rate_line: float = 0.026,  # 0.013,  # fails per year
    fail_rate_trafo: float = 0.0,  # fails per year
    ev_percentage: float = 0.46,
    ev_E_max: float = 0.07,
    random_seed: int = 2837314,
    iterations: int = 2,
    start_time: TimeStamp = TimeStamp(
        year=2019,
        month=1,
        day=1,
        hour=0,
        minute=0,
        second=0,
    ),
    stop_time: TimeStamp = TimeStamp(
        year=2020,
        month=1,
        day=1,
        hour=0,
        minute=0,
        second=0,
    ),
    save_iterations: list = [1, 2],
    data_dir: str = os.path.join(
        os.pardir,
        "load",
        "data",
    ),
    save_flag: bool = True,
    save_dir: str = "results",
    n_procs: int = 1,
    debug: bool = True,
):

    start = time.time()

    ps = initialize_network(
        include_microgrid=include_microgrid,
        include_production=include_production,
        include_ev=include_ev,
        v2g_flag=v2g_flag,
        line_repair_time_stat_dist=line_repair_time_stat_dist,
        fail_rate_line=fail_rate_line,  # fails per year
        fail_rate_trafo=fail_rate_trafo,
        ev_percentage=ev_percentage,
        ev_E_max=ev_E_max,
        include_ICT=False,
    )

    ps = set_network_load_and_prod(
        power_system=ps,
        include_microgrid=include_microgrid,
        include_production=include_production,
        data_dir=data_dir,
    )

    sim = Simulation(ps, random_seed=random_seed)

    sim.run_monte_carlo(
        iterations=iterations,
        start_time=start_time,
        stop_time=stop_time,
        time_step=Time(5, TimeUnit.MINUTE),
        time_unit=TimeUnit.HOUR,
        save_iterations=save_iterations,
        save_dir=save_dir,
        save_flag=save_flag,
        n_procs=n_procs,
        debug=debug,
    )

    end = time.time()
    print("Time elapsed: {}".format(end - start))



from abc import ABC, abstractmethod

from relsad.network.components import Bus, Line


class PowerNetwork(ABC):

    # Enforce bus-list as attribute
    buses: list = None

    # Enforce line-list as attribute
    lines: list = None

    # Enforce section-list as attribute
    sections: list = None

    @abstractmethod
    def add_bus(self, bus: Bus):
        pass

    @abstractmethod
    def add_buses(self, buses: list):
        pass

    @abstractmethod
    def add_line(self, line: Line):
        pass

    @abstractmethod
    def add_lines(self, lines: list):
        pass

    @abstractmethod
    def get_lines(self):
        pass

    @abstractmethod
    def reset_slack_bus(self):
        pass



import matplotlib.lines as mlines
import numpy as np

from relsad.Time import Time, TimeUnit
from relsad.utils import convert_yearly_fail_rate, random_choice

from .Component import Component
from .Controller import Controller, ControllerState
from .ICTNode import ICTNode


class MainController(Component, Controller):

    """
    Common class for main controller

    ...

    Attributes
    ----------
    name : string
        Name of the main controller
    ict_node : ICTNode
        The ICT node connected to the controller
    harware_fail_rate_per_year : float
        The failure rate per year for hardware failuers of the controller
    software_fail_rate_per_year : float
        The failure rate per year for software failures of the controller
    p_fail_repair_new_signal : float
        The probability tuat a new signal cannot be sent
    p_fail_reboot : float
        The probability that a reboot of the controller will fail to fix the failure
    new_signal_time : Time
        The time it takes to send a new signal
    reboot_time : Time
        The time it takes to reboot the controller
    remaining_repair_time : Time
        The remaining time of the repair
    manual_software_repair_time : Time
        The time it takes to manually repair a software failure
    manual_hardware_repair_time : Time
        The time it takes to manually repair a hardware failure
    state : enum
        The state of the controller
    sectioning_time : Time
        The sectioning time of the system (can differ based on if the controller is OK or not)
    manual_sectioning_time : Time
        The sectioning time of a passive system without controller (controller is failed)
    distribution_controllers : list
        List of controllers from the distribution systems in the power system
    microgrid_controllers : list
        List of controllers from the microgrids in the power system
    history : dict
        Dictonary attribute that stores the historic variables
    monte_carlo_history : dict
        Dictonary attribute that stores the historic variables from the Monte Carlo simulation


    Methods
    ----------
    fail_software()
        Sets the controller state to software fail
    fail_hardware()
        Sets the controller state to hardware fail
    not_fail()
        Sets the controller state to ok (not fail)
    draw_fail_status(dt)
        Draws the failure status of the controller
    draw_hardware_status(prob)
        Draws if a hardware failure has occured and sets the fail status of the controller
    draw_software_status(prob)
        Draws if a software failure has occured and sets the fail status of the controller
    repair_software_fail(dt)
        Goes through the repair sequence of a software failure, draws the repair mechanism, returns the repair time and updates the state of the controller
    update_fail_status(dt)
        Updates the failure status of the controller based on the remaining outage time
    add_distribution_controller(controller)
        Adds distribution controllers from connected distribution systems to a list and
        sets the sectioning time for the controller. Assignes the
        ICT node to the distribution system controller.
    add_microgrid_controller(controller)
        Adds microgird controllers from connected microgrids to a list and
        sets the sectioning time for the controller. Assignes the
        ICT node to the microgrid controller.
    run_control_loop(curr_time, dt)
        Runs the control loop for the distribution system controllers and the microgird controllers
    spread_sectioning_time_to_sub_controllers()
        Sets the sectioning time for the controllers in the connected distribution systems and microgrids
    update_history(prev_time, curr_time, save_flag)
        Updates the history variables
    get_history(attribute)
        Returns the history variables of an attribute
    add_random_instance(random_gen)
        Adds global random seed
    print_status()
        Prints the status
    reset_status(save_flag)
        Resets and sets the status of the class parameters
    initialize_history()
        Initializes the history variables
    """

    ## Random instance
    ps_random: np.random.Generator = None

    def __init__(
        self,
        name: str,
        ict_node: ICTNode = None,
        hardware_fail_rate_per_year: float = 0.2,
        software_fail_rate_per_year: float = 12,
        p_fail_repair_new_signal: float = 1 - 0.95,
        p_fail_repair_reboot: float = 0.9,
        new_signal_time: Time = Time(2, TimeUnit.SECOND),
        reboot_time: Time = Time(5, TimeUnit.MINUTE),
        manual_software_repair_time: Time = Time(0.3, TimeUnit.HOUR),
        manual_hardware_repair_time: Time = Time(2.5, TimeUnit.HOUR),
        manual_sectioning_time: Time = Time(1, TimeUnit.HOUR),
        state: ControllerState = ControllerState.OK,
    ):

        self.name = name
        self.ict_node = ict_node
        self.hardware_fail_rate_per_year = hardware_fail_rate_per_year
        self.software_fail_rate_per_year = software_fail_rate_per_year
        self.p_fail_repair_new_signal = p_fail_repair_new_signal
        self.p_fail_repair_reboot = p_fail_repair_reboot
        self.new_signal_time = new_signal_time
        self.reboot_time = reboot_time
        self.remaining_repair_time = Time(0)
        self.manual_software_repair_time = manual_software_repair_time
        self.manual_hardware_repair_time = manual_hardware_repair_time
        self.state = state
        self.sectioning_time = Time(0)
        self.manual_sectioning_time = manual_sectioning_time

        self.distribution_controllers = list()
        self.microgrid_controllers = list()

        ## History
        self.history = {}
        self.monte_carlo_history = {}
        self.initialize_history()

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"MainController(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name and isinstance(
                other, MainController
            )
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def fail_software(self):
        """
        Sets the controller state to software fail

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.state = ControllerState.SOFTWARE_FAIL

    def fail_hardware(self):
        """
        Sets the controller state to hardware fail

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.state = ControllerState.HARDWARE_FAIL

    def not_fail(self):
        """
        Sets the controller state to ok (not fail)

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.state = ControllerState.OK

    def draw_fail_status(self, dt: Time):
        """
        Draws the failure status of the controller

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        p_hardware_fail = convert_yearly_fail_rate(
            self.hardware_fail_rate_per_year, dt
        )
        self.draw_hardware_status(p_hardware_fail)
        if self.state == ControllerState.OK:
            p_software_fail = convert_yearly_fail_rate(
                self.software_fail_rate_per_year, dt
            )
            self.draw_software_status(p_software_fail)

    def draw_hardware_status(self, prob):
        """
        Draws if a hardware failure has occured and sets the fail status of the controller

        Parameters
        ----------
        prob : float
            The probability of a hardware failure

        Returns
        ----------
        None

        """
        if random_choice(self.ps_random, prob):
            self.fail_hardware()
        else:
            self.not_fail()

    def draw_software_status(self, prob):
        """
        Draws if a software failure has occured and sets the fail status of the controller

        Parameters
        ----------
        prob : float
            The probability of a software failure

        Returns
        ----------
        None

        """
        if random_choice(self.ps_random, prob):
            self.fail_software()
        else:
            self.not_fail()

    def repair_software_fail(self, dt: Time):
        """
        Goes through the repair sequence of a software failure, draws the repair mechanism, returns the repair time and updates the state of the controller

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        repair_time : Time
            The repair time of the software failure

        """

        repair_time = self.new_signal_time
        self.draw_software_status(self.p_fail_repair_new_signal)
        if self.state == ControllerState.OK:
            return repair_time
        elif self.state == ControllerState.SOFTWARE_FAIL:
            repair_time += self.reboot_time
            self.draw_software_status(self.p_fail_repair_reboot)
            if self.state == ControllerState.SOFTWARE_FAIL:
                self.remaining_repair_time = self.manual_software_repair_time
                self.state = ControllerState.REPAIR
            return repair_time

    def update_fail_status(self, dt: Time):
        """
        Updates the failure status of the controller based on the remaining outage time

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        if self.state == ControllerState.REPAIR:
            self.remaining_repair_time -= dt
            if self.remaining_repair_time <= Time(0):
                self.not_fail()
        elif self.state == ControllerState.OK:
            self.draw_fail_status(dt)
            if self.state == ControllerState.HARDWARE_FAIL:
                self.remaining_repair_time = self.manual_hardware_repair_time
                self.state = ControllerState.REPAIR
            elif self.state == ControllerState.SOFTWARE_FAIL:
                self.sectioning_time = self.repair_software_fail(dt)
                self.spread_sectioning_time_to_sub_controllers()

    def add_distribution_controller(self, controller):
        """
        Adds distribution controllers from connected distribution systems to a list and
        sets the sectioning time for the controller. Assignes the
        ICT node to the distribution system controller.

        Parameters
        ----------
        controller : Controller
            Distribution system controller

        Returns
        ----------
        None

        """
        self.distribution_controllers.append(controller)
        controller.manual_sectioning_time = self.manual_sectioning_time
        controller.ict_node = self.ict_node
        if self.ict_node is not None:
            controller.ict_network = self.ict_node.parent_network

    def add_microgrid_controller(self, controller):
        """
        Adds microgird controllers from connected microgrids to a list and
        sets the sectioning time for the controller. Assignes the
        ICT node to the microgrid controller.

        Parameters
        ----------
        controller : Controller
            Microgird controller

        Returns
        ----------
        None

        """
        self.microgrid_controllers.append(controller)
        controller.manual_sectioning_time = self.manual_sectioning_time
        controller.ict_node = self.ict_node
        if self.ict_node is not None:
            controller.ict_network = self.ict_node.parent_network

    def run_control_loop(self, curr_time: Time, dt: Time):
        """
        Runs the control loop for the distribution system controllers and the microgird controllers

        Parameters
        ----------
        curr_time : Time
            The current time
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        if self.state == ControllerState.OK:
            for controller in self.distribution_controllers:
                controller.run_control_loop(curr_time, dt)
            for controller in self.microgrid_controllers:
                controller.run_control_loop(curr_time, dt)
        elif self.state == ControllerState.REPAIR:
            for controller in self.distribution_controllers:
                controller.run_manual_control_loop(curr_time, dt)
            for controller in self.microgrid_controllers:
                controller.run_manual_control_loop(curr_time, dt)

    def spread_sectioning_time_to_sub_controllers(self):
        """
        Sets the sectioning time for the controllers in the connected distribution systems and microgrids

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        for controller in self.distribution_controllers:
            if controller.power_network.connected_line.circuitbreaker.is_open:
                controller.set_sectioning_time(self.sectioning_time)
        for controller in self.microgrid_controllers:
            if controller.power_network.connected_line.circuitbreaker.is_open:
                controller.set_sectioning_time(self.sectioning_time)

    def update_history(
        self, prev_time: Time, curr_time: Time, save_flag: bool
    ):
        """
        Updates the history variables

        Parameters
        ----------
        prev_time : Time
            The previous time
        curr_time : Time
            Current time
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        if save_flag:
            time = curr_time.get_unit_quantity(curr_time.unit)
            self.history["sectioning_time"][
                time
            ] = self.sectioning_time.get_unit_quantity(curr_time.unit)
            self.history["remaining_repair_time"][
                time
            ] = self.remaining_repair_time.get_unit_quantity(curr_time.unit)
            self.history["state"][time] = self.state.value

    def get_history(self, attribute: str):
        """
        Returns the history variables of an attribute

        Parameters
        ----------
        attribute : str
            Controller attribute

        Returns
        ----------
        history[attribute] : dict
            Returns the history variables of an attribute

        """
        return self.history[attribute]

    def add_random_instance(self, random_gen):
        """
        Adds global random seed

        Parameters
        ----------
        random_gen : int
            Random number generator

        Returns
        ----------
        None

        """
        self.ps_random = random_gen

    def print_status(self):
        """
        Prints the status

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        pass

    def reset_status(self, save_flag: bool):
        """
        Resets and sets the status of the class parameters

        Parameters
        ----------
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        self.state = ControllerState.OK
        self.sectioning_time = Time(0)
        self.remaining_repair_time = Time(0)
        if save_flag:
            self.initialize_history()

    def initialize_history(self):
        """
        Initializes the history variables

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.history["sectioning_time"] = {}
        self.history["remaining_repair_time"] = {}
        self.history["state"] = {}



import matplotlib.pyplot as plt

from relsad.network.components import (
    Bus,
    CircuitBreaker,
    Disconnector,
    EVPark,
    IntelligentSwitch,
    Line,
    Sensor,
    ICTNode,
    ICTLine,
)


def plot_topology(
    buses: list,
    lines: list,
    ict_nodes: list = None,
    ict_lines: list = None,
    bus_text: bool = True,
    line_text: bool = False,
    circuitbreaker_text: bool = False,
    disconnector_text: bool = False,
    intelligent_switch_text: bool = False,
    sensor_text: bool = False,
    text_dx: tuple = (0, -0.1),
    **kwargs
):

    """
    Plots the system topology

    Parameters
    ----------
    buses : list
        List with Bus elements in the system
    lines : list
        List with Line elements in the system
    bus_text : bool
        Flag determining if bus name will be plotted
    line_text : bool
        Flag determining if line name will be plotted
    circuitbreaker_text : bool
        Flag determining if circuitbreaker name will be plotted
    disconnector_text : bool
        Flag determining if disconnector name will be plotted
    intelligent_switch_text : bool
        Flag determining if intelligent switch name will be plotted
    sensor_text : bool
        Flag determining if sensor name will be plotted
    **kwargs : dict
        Plotting keyword arguments.

    Returns
    ----------
    fig : figure
        Figure of the system topology
    None

    """
    # Handle keyword arguments

    left = kwargs["left"] if "left" in kwargs else 0.02
    right = kwargs["right"] if "right" in kwargs else 0.98
    bottom = kwargs["bottom"] if "bottom" in kwargs else 0.1
    top = kwargs["top"] if "top" in kwargs else 0.75

    text_size = kwargs["text_size"] if "text_size" in kwargs else 6

    ncol = kwargs["ncol"] if "ncol" in kwargs else None

    # Clear used keyword arguments
    kwargs.pop("left", None)
    kwargs.pop("right", None)
    kwargs.pop("bottom", None)
    kwargs.pop("top", None)
    kwargs.pop("text_size", None)
    kwargs.pop("ncol", None)

    fig, ax = plt.subplots(**kwargs)

    fig.subplots_adjust(
        left=left,
        bottom=bottom,
        right=right,
        top=top,
        wspace=None,
        hspace=None,
    )
    legends = {}
    for bus in buses:
        _plot_bus(
            ax=ax,
            bus=bus,
            text=bus_text,
            text_size=text_size,
            text_dx=text_dx,
        )
        legends["Bus"] = Bus.handle
        if bus.ev_park is not None:
            legends["EV park"] = EVPark.handle
    for line in lines:
        _plot_line(
            ax=ax,
            line=line,
            text=line_text,
            text_size=text_size,
        )
        legends["Line"] = Line.handle
        if line.circuitbreaker is not None:
            _plot_circuitbreaker(
                ax=ax,
                line=line,
                text=circuitbreaker_text,
                text_size=text_size,
            )
            legends["Circuit breaker"] = CircuitBreaker.handle
        for discon in line.disconnectors:
            _plot_disconnector(
                ax=ax,
                discon=discon,
                text=disconnector_text,
                text_size=text_size,
            )
            legends["Disconnector"] = Disconnector.handle
            if discon.intelligent_switch:
                _plot_intelligent_switch(
                    ax=ax,
                    discon=discon,
                    text=intelligent_switch_text,
                    text_size=text_size,
                )
                legends["Intelligent switch"] = IntelligentSwitch.handle
        if line.sensor:
            _plot_sensor(
                ax=ax,
                sensor=line.sensor,
                text=sensor_text,
                text_size=text_size,
            )
            legends["Sensor"] = Sensor.handle
    if ict_nodes is not None:
        for ict_node in ict_nodes:
            _plot_ict_node(
                ax=ax,
                ict_node=ict_node,
                text=False,
                text_size=text_size,
                text_dx=text_dx,
            )
            legends["ICT node"] = ICTNode.handle
    if ict_lines is not None:
        for ict_line in ict_lines:
            _plot_ict_line(
                ax=ax,
                ict_line=ict_line,
                text=False,
                text_size=text_size,
            )
            legends["ICT line"] = ICTLine.handle

    plt.figlegend(
        legends.values(),
        legends.keys(),
        ncol=ncol if ncol is not None else len(legends),
        loc="upper center",
        bbox_to_anchor=(left + (right - left) / 2, 0.978),
        frameon=False,
        prop={"size": 8},
        handleheight=3,
    )

    plt.axis("off")

    return fig


def _plot_bus(
    ax: plt.axis,
    bus: Bus,
    text: bool = False,
    text_size: int = 8,
    text_dx: tuple = (0, 0),
):
    """
    Plot bus

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Plot axis
    bus : Bus
        Bus element
    text : bool
        Flag determining if bus name will be plotted
    text_size : int
        The size of the text in the plot
    text_dx : tuple
        The distance from the text to the center point

    Returns
    ----------
    None

    """
    x, y = bus.coordinate
    ax.plot(
        x,
        y,
        marker=bus.marker,
        markeredgewidth=bus.handle.get_markeredgewidth(),
        markersize=bus.size,
        linestyle="None",
        color=bus.color,
        clip_on=False,
        zorder=3,
    )
    if text:
        ax.text(
            x + text_dx[0],
            y + text_dx[1],
            bus.name,
            ha="center",
            va="center",
            size=text_size,
        )
    if bus.ev_park is not None:
        ax.plot(
            x,
            y,
            marker=bus.ev_park.marker,
            markeredgewidth=bus.ev_park.handle.get_markeredgewidth(),
            markersize=bus.ev_park.size,
            linestyle="None",
            color=bus.color,
            clip_on=False,
            zorder=3,
        )

def _plot_line(
    ax: plt.axis,
    line: Line,
    text: bool = False,
    text_size: int = 8,
):
    """
    Plot line

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Plot axis
    line : Line
        A Line element
    text : bool
        Flag determining if line name will be plotted
    text_size : int
        The size of the text in the plot

    Returns
    ----------
    None

    """
    dx_fraction = 0.1

    x1, y1 = line.fbus.coordinate
    x2, y2 = line.tbus.coordinate

    linestyle = ":" if line.is_backup else line.linestyle
    ax.plot(
        [x1, x2],
        [y1, y2],
        color=line.color,
        linestyle=linestyle,
        zorder=2,
    )
    if text:
        dx = x2 - x1
        dy = y2 - y1
        x = (x1 + x2) / 2
        y = (y1 + y2) / 2
        if dx == 0 and dy != 0:
            x += dy * dx_fraction
        elif dx != 0 and dy == 0:
            y += dx * dx_fraction
        else:
            x += dx * dx_fraction
            y -= dy * dx_fraction
        ax.text(
            x,
            y,
            line.name,
            ha="center",
            va="center",
            size=text_size,
        )


def _plot_circuitbreaker(
    ax: plt.axis,
    line: Line,
    text: bool = False,
    text_size: int = 8,
):
    """
    Plot circuitbreakers

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Plot axis
    line : Line
        A Line element
    text : bool
        Flag determining if circuitbreaker name will be plotted
    text_size : int
        The size of the text in the plot

    Returns
    ----------
    None

    """
    cb = line.circuitbreaker
    x, y = cb.coordinate

    ax.plot(
        x,
        y,
        marker=cb.marker,
        markeredgewidth=cb.handle.get_markeredgewidth(),
        markersize=cb.size,
        linestyle="None",
        color=cb.color,
        markeredgecolor=cb.edgecolor,
        zorder=3,
    )
    if text:
        ax.text(
            x,
            y - 0.2,
            cb.name,
            ha="center",
            va="center",
            size=text_size,
        )


def _plot_disconnector(
    ax: plt.axis,
    discon: Disconnector,
    text: bool = False,
    text_size: int = 8,
):
    """
    Plot disconnectors

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Plot axis
    discon : Disconnecotr
        A Disconnector elements
    text : bool
        Flag determining if disconnector name will be plotted
    text_size : int
        The size of the text in the plot

    Returns
    ----------
    None

    """
    x, y = discon.coordinate
    ax.plot(
        x,
        y,
        marker=discon.marker,
        markeredgewidth=discon.handle.get_markeredgewidth(),
        markersize=discon.size,
        linestyle="None",
        color=discon.color,
        markeredgecolor=discon.edgecolor,
        zorder=3,
    )
    if text:
        ax.text(
            x,
            y - 0.2,
            discon.name,
            ha="center",
            va="center",
            size=text_size,
        )


def _plot_intelligent_switch(
    ax: plt.axis,
    discon: Disconnector,
    text: bool = False,
    text_size: int = 8,
):
    """
    Plot intelligent switches

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Plot axis
    discon : Disconnector
        A Disconnector elements
    text : bool
        Flag determining if intelligent switch name will be plotted
    text_size : int
        The size of the text on the plot

    Returns
    ----------
    None

    """
    x, y = discon.coordinate
    ax.plot(
        x,
        y,
        marker=discon.intelligent_switch.marker,
        markeredgewidth=discon.intelligent_switch.handle.get_markeredgewidth(),
        markersize=discon.intelligent_switch.size,
        linestyle="None",
        color=discon.intelligent_switch.color,
        zorder=3,
    )
    if text:
        ax.text(
            x,
            y - 0.2,
            discon.intelligent_switch.name,
            ha="center",
            va="center",
            size=text_size,
        )


def _plot_sensor(
    ax: plt.axis,
    sensor: Sensor,
    text: bool = False,
    text_size: int = 8,
):
    """
    Plot sensors

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Plot axis
    sensor : Sensor
        A Line element
    text : bool
        Flag determining if sensor name will be plotted
    text_size : int
        The size of the text in the plot

    Returns
    ----------
    None

    """
    x, y = sensor.coordinate

    ax.plot(
        x,
        y,
        marker=sensor.marker,
        markeredgewidth=sensor.handle.get_markeredgewidth(),
        markersize=sensor.size,
        linestyle="None",
        color=sensor.color,
        zorder=3,
    )
    if text:
        ax.text(
            x,
            y + 0.2,
            sensor.name,
            ha="center",
            va="center",
            size=text_size,
        )

def _plot_ict_node(
    ax: plt.axis,
    ict_node: ICTNode,
    text: bool = False,
    text_size: int = 8,
    text_dx: tuple = (0, 0),
):
    """
    Plot ICT node

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Plot axis
    ict_node : ICTNode
        ICTNode element
    text : bool
        Flag determining if bus name will be plotted
    text_size : int
        The size of the text in the plot
    text_dx : tuple
        The distance from the text to the center point

    Returns
    ----------
    None

    """
    x, y = ict_node.coordinate
    ax.plot(
        x,
        y,
        marker=ict_node.marker,
        markeredgewidth=ict_node.handle.get_markeredgewidth(),
        markersize=ict_node.size,
        linestyle="None",
        color=ict_node.color,
        clip_on=False,
        zorder=3,
    )
    if text:
        ax.text(
            x + text_dx[0],
            y + text_dx[1],
            ict_node.name,
            ha="center",
            va="center",
            size=text_size,
        )


def _plot_ict_line(
    ax: plt.axis,
    ict_line: ICTLine,
    text: bool = False,
    text_size: int = 8,
):
    """
    Plot ICT line

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Plot axis
    ict_line : ICTLine
        An ICTLine element
    text : bool
        Flag determining if line name will be plotted
    text_size : int
        The size of the text in the plot

    Returns
    ----------
    None

    """
    dx_fraction = 0.1

    x1, y1 = ict_line.fnode.coordinate
    x2, y2 = ict_line.tnode.coordinate

    ax.plot(
        [x1, x2],
        [y1, y2],
        color=ict_line.color,
        linestyle=ict_line.linestyle,
        zorder=2,
    )
    if text:
        dx = x2 - x1
        dy = y2 - y1
        x = (x1 + x2) / 2
        y = (y1 + y2) / 2
        if dx == 0 and dy != 0:
            x += dy * dx_fraction
        elif dx != 0 and dy == 0:
            y += dx * dx_fraction
        else:
            x += dx * dx_fraction
            y -= dy * dx_fraction
        ax.text(
            x,
            y,
            ict_line.name,
            ha="center",
            va="center",
            size=text_size,
        )



import os
import sys

import numpy as np

from relsad.examples.load.load_and_gen_data import (
    load_data,
    pv_power,
    weather_generation_data,
    wind_power,
)
from relsad.load.bus import CostFunction
from relsad.network.systems import PowerSystem
from relsad.Time import Time, TimeStamp, TimeUnit


def set_network_load_and_prod(
    power_system: PowerSystem,
):

    # Fetching bus-objects

    B1 = power_system.get_comp("B1")
    B2 = power_system.get_comp("B2")
    B3 = power_system.get_comp("B3")
    B4 = power_system.get_comp("B4")
    B5 = power_system.get_comp("B5")

    B7 = power_system.get_comp("B7")
    B8 = power_system.get_comp("B8")
    B9 = power_system.get_comp("B9")
    B10 = power_system.get_comp("B10")
    B11 = power_system.get_comp("B11")
    B12 = power_system.get_comp("B12")
    B13 = power_system.get_comp("B13")
    B14 = power_system.get_comp("B14")
    B15 = power_system.get_comp("B15")
    B16 = power_system.get_comp("B16")
    B17 = power_system.get_comp("B17")
    B18 = power_system.get_comp("B18")
    B19 = power_system.get_comp("B19")
    B20 = power_system.get_comp("B20")
    B21 = power_system.get_comp("B21")
    B22 = power_system.get_comp("B22")

    load_res1 = np.ones(8760) * (0.535 / 210)
    load_res2 = np.ones(8760) * (0.450 / 200)
    load_small1 = np.ones(8760) * (1 / 1)
    load_small2 = np.ones(8760) * (1.15 / 1)
    load_gov = np.ones(8760) * (0.566 / 1)
    load_com = np.ones(8760) * (0.454 / 10)

    household = CostFunction(
        A=8.8,
        B=14.7,
    )

    for bus in [B1, B2, B3, B10, B11]:
        bus.add_load_data(
            pload_data=load_res1,
            cost_function=household,
        )
    for bus in [B4, B5, B13, B14, B20, B21]:
        bus.add_load_data(
            pload_data=load_gov,
            cost_function=household,
        )
    for bus in [B5, B7, B15, B16, B22]:
        bus.add_load_data(
            pload_data=load_com,
            cost_function=household,
        )
    for bus in [B12, B17, B18, B19]:
        bus.add_load_data(
            pload_data=load_res2,
            cost_function=household,
        )
    B8.add_load_data(
        pload_data=load_small1,
        cost_function=household,
    )
    B9.add_load_data(
        pload_data=load_small2,
        cost_function=household,
    )

    return power_system



from relsad.examples.tutorial.system import *

from relsad.simulation import Simulation

sim = Simulation(power_system=ps, random_seed=0)
sim.run_monte_carlo(
    iterations=10,
    start_time=TimeStamp(
        year=2019,
        month=0,
        day=0,
        hour=0,
        minute=0,
        second=0,
    ),
    stop_time=TimeStamp(
        year=2020,
        month=0,
        day=0,
        hour=0,
        minute=0,
        second=0,
    ),
    time_step=Time(1, TimeUnit.HOUR),
    time_unit=TimeUnit.HOUR,
    callback=None,
    save_iterations=[1, 2],
    save_dir="results",
    n_procs=1,
)

path = os.path.join(
    "results",
    "monte_carlo",
    "ps1",
    "ENS.csv",
)

df = pd.read_csv(path, index_col=0)
fig, ax = plt.subplots()
df.hist(ax=ax)

fig.savefig(
    "ENS.png",
    dpi=600,
)

print(df.describe())

import shutil

shutil.rmtree("results")
os.remove("ENS.png")



from .StatDist import (
    StatDist,
    StatDistType,
    GammaParameters,
    NormalParameters,
    UniformParameters,
)

__all__ = []
for v in dir():
    if not v.startswith("__") and v != "relsad":
        __all__.append(v)



from .shedding import (
    shed_energy,
    _build_A_matrix,
    _get_generation_bounds,
    _gather_bounds,
    _shed_active_loads,
    _shed_active_energy,
    _shed_reactive_loads,
    _shed_reactive_energy,
)

__all__ = []
for v in dir():
    if not v.startswith("__") and v != "relsad":
        __all__.append(v)



from relsad.network.components import ICTLine, ICTNode
from relsad.utils import unique

from .PowerSystem import PowerSystem


class ICTNetwork:

    """
    Class defining a ICT network type

    ...

    Attributes
    ----------
    name : str
        Name of the ICT network
    nodes : list
        List containing the nodes connected to the ICT network
    lines : list
        List containing the lines connected to the ICT network
    comp_list : list
        List containing the components in the ICT network
    comp_dict : dict
        Dictionary containing the components in the ICT network
    power_system : Network
        Connects the ICT network to the power system
    failed_line : Bool
        Flag stating if the ICT network contains a failed line
    history : dict
        Dictionary containing the history variables of the network
    monte_carlo_history : dict
        Dictionary containing the history variables from the monte carlo simulation



    Methods
    ----------
    add_node(node)
        Adding a node to the ICT network
    add_nodes(nodes)
        Adding nodes to the ICT network
    add_line(line)
        Adding a line to the ICT network
    add_lines(lines)
        Adding lines to the ICT network
    get_lines()
        Returns the lines in the ICT network
    get_monte_carlo_history(attribute)
        Returns the specified history variable from the Monte Carlo simulation
    get_history(attribute)
        Returns the specified history variable
    """

    ## Visual attributes
    color = "steelblue"

    ## Counter
    counter = 0

    def __init__(self, power_system: PowerSystem):
        ICTNetwork.counter += 1
        self.name = "ict_network{:d}".format(ICTNetwork.counter)

        # Components
        self.nodes = list()
        self.lines = list()
        self.comp_list = list()
        self.comp_dict = dict()

        # Network connections
        self.power_system = power_system

        self.failed_line = False

        ## History
        self.history: dict = {}
        self.monte_carlo_history: dict = {}

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"ICTNetwork(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def add_node(self, node: ICTNode):
        """
        Adding a node to the ICT network

        Parameters
        ----------
        node : ICTNode
            A ICT node element

        Returns
        ----------
        None

        """
        # Add bus to distribution network
        self.comp_dict[node.name] = node
        node.handle.color = self.color
        node.color = self.color
        node.parent_network = self
        self.nodes.append(node)
        self.nodes = unique(self.nodes)

        # Add bus to the power system
        self.power_system.add_ict_node(node)

    def add_nodes(self, nodes: list):
        """
        Adding buses to the ICT network node list

        Parameters
        ----------
        nodes : list
            A list of ICT node elements in the ICT network

        Returns
        ----------
        None

        """
        for node in nodes:
            self.add_node(node)

    def add_line(self, line: ICTLine):
        """
        Adding a line to the ICT network

        Parameters
        ----------
        line : ICTLine
            A ICTLine element

        Returns
        ----------
        None

        """

        # Add line to the ICT network
        line.handle.color = self.color
        line.color = self.color
        line.parent_network = self
        self.comp_dict[line.name] = line
        self.lines.append(line)
        self.lines = unique(self.lines)

        # Add line to power system
        self.power_system.add_ict_line(line)

    def add_lines(self, lines: list):
        """
        Adding lines to ICT network line list

        Parameters
        ----------
        lines : list
            A list of ICTLine elements in the ICT network

        Returns
        ----------
        None

        """
        for line in lines:
            self.add_line(line)

    def get_lines(self):
        """
        Returns the lines in the ICT network

        Parameters
        ----------
        None

        Returns
        ----------
        lines : list
            List of ICTLine elements

        """
        return self.lines

    def get_monte_carlo_history(self, attribute):
        """
        Returns the specified history variable from the Monte Carlo simulation

        Parameters
        ----------
        attribute : str
            ICT network attribute

        Returns
        ----------
        monte_carlo_history[attribute] : dict
            Returns the history variables of an attribute from the Monte Carlo simulation

        """
        return self.monte_carlo_history[attribute]

    def get_history(self, attribute):
        """
        Returns the specified history variable

        Parameters
        ----------
        attribute : str
            ICT network attribute

        Returns
        ----------
        history[attribute] : dict
            Returns the history variables of an attribute

        """
        return self.history[attribute]



from enum import Enum
from numbers import Number

import numpy as np

from relsad.Time import Time, TimeUnit
from relsad.utils import INF

from .Bus import Bus
from .Component import Component
from .MicrogridController import MicrogridMode


class BatteryType(Enum):
    """
    Battery type

    Attributes
    ----------
    REGULAR : int
        Regular battery type
    EV : int
        Electric vehicle battery type
    """

    REGULAR = 1
    EV = 2


class BatteryState(Enum):
    """
    Battery state

    Attributes
    ----------
    ACTIVE : int
        Battery is active
    INACTIVE : int
        Battery is inactive
    """

    ACTIVE = 1
    INACTIVE = 2


class Battery(Component):
    """
    Common class for batteries

    ...

    Attributes
    ----------
    name : string
        Name of the battery
    mode : str
        Microgrid mode
    survival_time : Time
        Total amount of hours the microgrid should survive on battery capacity
    remaining_survival_time : Time
        The time left for the battery to ensure energy for the microgrid load
    standard_SOC_min : float
        The minimum State of Charge for the battery which can change based on wanted battery capacity
    bus : Bus
        The bus the battery is connected to
    inj_p_max : float
        The maximum active power that the battery can inject [MW]
    inj_q_max : float
        The maximum reactive power that the battery can inject [MVar]
    inj_max : float
        The maximum apperent power that the battery can inject [MVa]
    f_inj_p : float
        Active power capacity fraction
    f_inj_q : float
        Reactive power capacity fraction
    E_max : float
        The maximum capacity of the battery [MWh]
    SOC_min : float
        The minimal state of charge level in the battery
    SOC_max : float
        The maximum state of charge level in the battery
    n_battery : float
        The battery efficiency
    SOC_start : float
        The iteration start value for the state of charge level in the battery
    p_inj : float
        The injected active power in the battery [MW]
    q_inj : float
        The injected reactive power in the battery [MVar]
    SOC : float
        The state of charge of the battery
    E_battery : float
        The amount of energy stored in the battery [MWh]
    battery_type : BatteryType
        The battery type
    state : BatteryState
        Battery state
    history : dict
        Dictonary attribute that stores the historic variables

    Methods
    ----------
    update_SOC()
        Updates the SOC in the battery
    charge(P_ch, dt)
        Charge the battery. Decides how much the battery can charge based on the desired charging power. Restricted by the amount of power that can be stored, the maximum power that can be injected, and the maximum state of charge of the battery. Updates the state of charge of the battery.
        Returns a float telling how much power the battery is not able to charge
    discharge(P_dis, q_dis, dt)
        Discharge the battery. Decides how much the battery can discharge based on the available energy in the battery. Limited by the state of charge, the maximum power that can be injected, and the wanted amount of power from the battery. Updates the state of charge of the battery.
        Returns a float telling how much power the battery is not able to discharge
    print_status()
        Prints the status of the battery
    update_bus_load_and_prod(system_load_balance_p, system_load_balance_q)
        Updates the load and production on the bus based on the system load balance.
        If the balance is negative, there is a surplus of production, and the battery will charge.
        If the balance is positive, there is a shortage of production, and the battery will discharge.
        Returns the remaining surplus/shortage of power
    initialize_history()
        Initializes the history variables
    update_history(prev_time, curr_time, dt, save_flag)
        Updates the history variables
    get_history(attribute)
        Returns the history variables of an attribute
    update_fail_status(dt)
        Locks and unlocks the battery functionality based on failure states of the basestation
    add_random_instance(random_gen)
        Adds global random seed
    reset_status()
        Resets and sets the status of the system parameters
    set_mode(mode)
        Sets the microgrid mode
    start_survival_time()
        Starts the timer for how long the battery should focus on supporting own load
        Only for when a microgrid is added and follows a survival mode
    set_SOC_state()
        Sets the SOC state
    draw_SOC_state()
        Draws the SOC state based on a uniform distribution
    update(p, q, fail_duration, dt)
        Updates the battery status for the current time step

    """

    ## Random instance
    ps_random: np.random.Generator = None

    def __init__(
        self,
        name: str,
        bus: Bus,
        inj_p_max: float = 0.5,
        inj_q_max: float = 0.5,
        E_max: float = 1,
        SOC_min: float = 0.1,
        SOC_max: float = 1,
        n_battery: float = 0.95,
        battery_type: BatteryType = BatteryType.REGULAR,
        random_instance: np.random.Generator = None,
        SOC_start: float = None,
    ):

        # Verify input
        if bus is None:
            raise Exception("Battery must be connected to a Bus")
        if bus.parent_network is not None and random_instance is None:
            raise Exception(
                "Battery must be created before the bus is connected to a network"
            )
        if inj_p_max < 0:
            raise Exception("The active power injection must be positive")
        if inj_q_max < 0:
            raise Exception("The reactive power injection must be positive")
        if E_max < 0:
            raise Exception("The energy capacity must be positive")
        if SOC_min < 0 or SOC_max > 1:
            raise Exception("The SOC limits must be between 0 and 1")
        if n_battery < 0 or n_battery > 1:
            raise Exception("The efficiency must be between 0 and 1")
        if (
            SOC_start is not None
            and not isinstance(SOC_start, Number)
            and (SOC_start < 0 or SOC_start > 1)
        ):
            raise Exception(
                "The SOC start value must be a number between 0 and 1"
            )

        self.name = name

        self.ps_random = random_instance

        self.mode = None
        self.survival_time = Time(4, TimeUnit.HOUR)
        self.remaining_survival_time = Time(0)
        self.standard_SOC_min = SOC_min

        self.bus = bus

        self.battery_type = battery_type

        if battery_type == BatteryType.REGULAR:
            # Link battery to parent bus
            bus.battery = self

        self.inj_p_max = inj_p_max  # MW
        self.inj_q_max = inj_q_max  # MVar
        self.inj_max = self.inj_p_max
        # active capacity fraction
        self.f_inj_p = 1
        # reactive capacity fraction
        self.f_inj_q = self.inj_q_max / self.inj_max

        self.E_max = E_max  # MWh
        self.SOC_min = SOC_min
        self.E_min = E_max * SOC_min  # MWh
        self.SOC_max = SOC_max
        self.n_battery = n_battery

        if SOC_start is None:
            self.set_SOC_state(SOC_state=SOC_min)
        else:
            self.set_SOC_state(SOC_state=SOC_start)
        self.SOC_start = SOC_start

        self.p_inj = 0.0  # MW
        self.q_inj = 0.0  # MVar

        self.state = BatteryState.ACTIVE

        ## History
        self.history = {}
        self.initialize_history()

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"Battery(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name and isinstance(other, Battery)
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def update_SOC(self):
        """
        Updates the SOC in the battery

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.SOC = self.E_battery / self.E_max

    def charge(self, p_ch, dt: Time):

        """
        Charge the battery

        Decides how much the battery can charge based on the desired charging power. Restricted by the amount of power that can be stored, the maximum power that can be injected, and the maximum state of charge of the battery

        Updates the state of charge of the battery

        Returns a float telling how much power the battery is not able to charge


        Parameters
        ----------
        p_ch : float
            Desired amount of active power to charge the battery  [MW]
        dt : Time
            The current time step

        Returns
        ----------
        P_ch_remaining : float
            Amount of active power exceeding the charging capacity [MW]


        """

        p_ch_remaining = 0
        if p_ch > self.inj_p_max:
            p_ch_remaining += p_ch - self.inj_p_max
            if p_ch >= INF:
                # "Infinite" power source available
                p_ch = self.inj_p_max
            else:
                p_ch -= p_ch_remaining
        # Change in energy
        dE = self.n_battery * p_ch * dt.get_hours()  # MWh
        # Energy level, trial step
        E_tr = self.E_battery + dE
        SOC_tr = E_tr / self.E_max
        dSOC = dE / self.E_max
        if SOC_tr > self.SOC_max:
            # Correcting step
            #
            # Factor to scale the energy level
            f = 1 - (SOC_tr - self.SOC_max) / dSOC
            self.E_battery += f * dE
            p_ch_remaining += (1 - f) * p_ch
        else:
            self.E_battery += dE
            p_ch_remaining += 0
        self.update_SOC()
        return p_ch_remaining

    def discharge(self, p_dis, q_dis, dt: Time):

        """
        Discharge the battery

        Decides how much the battery can discharge based on the available energy in the battery. Limited by the state of charge, the maximum power that can be injected, and the wanted amount of power from the battery

        Updates the state of charge of the battery

        Returns a float telling how much power the battery is not able to discharge


        Parameters
        ----------
        p_dis : float
            Amount of wanted active power from the network [MW]
        q_dis : float
            Amount of wanted reactive power from the network [MVar]
        dt : Time
            The current time step

        Returns
        ----------
        p_dis_remaining : float
            Amount of wanted active power from the network exceeding battery discharging capacity [MW]
        q_dis_remaining : float
            Amount of wanted reactive power from the network exceeding battery discharging capacity [MVar]

        """

        if (
            self.remaining_survival_time > Time(0)
            and self.mode == MicrogridMode.SURVIVAL
        ):
            self.remaining_survival_time -= dt
            self.SOC_min = min(
                self.bus.parent_network.get_max_load()[0]
                * self.remaining_survival_time.get_hours()
                / self.E_max
                + self.standard_SOC_min,
                self.SOC_max,
            )
        else:
            self.SOC_min = self.standard_SOC_min

        p_dis_remaining = 0
        q_dis_remaining = 0
        if p_dis > self.inj_p_max:
            p_dis_remaining += p_dis - self.inj_p_max
            p_dis -= p_dis_remaining
        if q_dis > self.inj_q_max:
            q_dis_remaining += q_dis - self.inj_q_max
            q_dis -= q_dis_remaining
        if p_dis + q_dis > self.inj_max:
            f_p = p_dis / (p_dis + q_dis)  # active fraction
            f_q = 1 - f_p  # reactive fraction
            diff = p_dis + q_dis - self.inj_max
            p_dis_remaining += diff * (1 - f_p)
            q_dis_remaining += diff * (1 - f_q)
            p_dis -= diff * (1 - f_p)
            q_dis -= diff * (1 - f_q)

        dE = 1 / self.n_battery * (p_dis + q_dis) * dt.get_hours()  # MWh/MVarh
        E_tr = self.E_battery - dE
        SOC_tr = E_tr / self.E_max
        dSOC = dE / self.E_max
        if SOC_tr < self.SOC_min and dSOC > 0:
            f = 1 - (self.SOC_min - SOC_tr) / dSOC
            self.E_battery -= f * dE
            p_dis_remaining += (1 - f) * p_dis
            q_dis_remaining += (1 - f) * q_dis
        else:
            self.E_battery -= dE
        self.update_SOC()
        return p_dis_remaining, q_dis_remaining

    def print_status(self):

        """
        Prints the status of the battery

        Parameters
        ----------
        None

        Returns
        ----------
        None
        """

        print("Battery status")
        print("name: {}".format(self.name))
        print("parent bus: {}".format(self.bus.name))
        print("inj_p_max: {} MW".format(self.inj_p_max))
        print("inj_q_max: {} MVar".format(self.inj_q_max))
        print("E_max: {} MWh".format(self.E_max))
        print("Efficiency: {} %".format(self.n_battery * 100))
        print("SOC_min: {}".format(self.SOC_min))
        print("SOC_max: {}".format(self.SOC_max))
        print("SOC: {:.2f}".format(self.SOC))

    def update_bus_load_and_prod(
        self,
        system_load_balance_p: float,
        system_load_balance_q: float,
        dt: Time,
    ):
        """
        Updates the load and production on the bus based on the system load balance.
        If the balance is negative, there is a surplus of production, and the battery will charge.
        If the balance is positive, there is a shortage of production, and the battery will discharge.
        Returns the remaining surplus/shortage of power

        Parameters
        ----------
        system_load_balance_p : float
            Active power system load balance in MW
        system_load_balance_q : float
            Reactive power system load balance in MW
        dt : Time
            The current time step

        Returns
        ----------
        p_rem : float
            Remaining surplus/shortage active power
        q_rem : float
            Remianing surplus/shortage reactive power
        """
        p, q = system_load_balance_p, system_load_balance_q

        if self.state == BatteryState.INACTIVE:  # Trafo has failed
            p_rem, q_rem = p, q
            return p_rem, q_rem

        pprod, qprod, pload, qload = 0, 0, 0, 0
        if p >= 0 and q >= 0:
            p_rem, q_rem = self.discharge(p, q, dt)
            pprod = p - p_rem
            qprod = q - q_rem
        if p < 0 and q >= 0:
            p_rem = self.charge(-p, dt)
            q_rem = self.discharge(0, q, dt)[1]
            pload = -p - p_rem
            qprod = q - q_rem
        if p >= 0 and q < 0:
            p_rem = self.discharge(p, 0, dt)[0]
            pprod = p - p_rem
        if p < 0 and q < 0:
            p_rem = self.charge(-p, dt)
            pload = -p - p_rem

        # Add production to bus
        self.bus.pprod += pprod  # MW
        self.bus.qprod += qprod  # MVar
        self.bus.pprod_pu += pprod / self.bus.s_ref  # PU
        self.bus.qprod_pu += qprod / self.bus.s_ref  # PU

        # Add load to bus
        self.bus.add_load(
            pload=pload,
            qload=qload,
        )

        # Battery power injection
        self.p_inj = pload - pprod
        self.q_inj = qload - qprod

        # Remaining power
        p_rem = p + pload - pprod
        q_rem = q + qload - qprod
        return p_rem, q_rem

    def initialize_history(self):
        """
        Initializes the storage of the history variables

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.history["SOC"] = {}
        self.history["SOC_min"] = {}
        self.history["remaining_survival_time"] = {}

    def update_history(
        self, prev_time: Time, curr_time: Time, save_flag: bool
    ):
        """
        Updates the history variables

        Parameters
        ----------
        prev_time : Time
            The previous time
        curr_time : Time
            Current time
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        if save_flag:
            time = curr_time.get_unit_quantity(curr_time.unit)
            self.history["SOC"][time] = self.SOC
            self.history["SOC_min"][time] = self.SOC_min
            self.history["remaining_survival_time"][
                time
            ] = self.remaining_survival_time.get_unit_quantity(curr_time.unit)

    def get_history(self, attribute: str):
        """
        Returns the history variables of an attribute

        Parameters
        ----------
        attribute : str
            Battery attribute

        Returns
        ----------
        history[attribute] : dict
            Returns the history variables of an attribute

        """
        return self.history[attribute]

    def update_fail_status(self, dt: Time):
        """
        Locks og unlocks the battery functionality based on failure states of the basestation

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        if self.bus.trafo_failed:
            self.state = BatteryState.INACTIVE
        else:
            self.state = BatteryState.ACTIVE

    def add_random_instance(self, random_gen):
        """
        Adds global random seed

        Parameters
        ----------
        random_gen : int
            Random number generator

        Returns
        ----------
        None

        """
        self.ps_random = random_gen

    def reset_status(self, save_flag: bool):
        """
        Resets and sets the status of the class parameters

        Parameters
        ----------
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        if self.SOC_start is None:
            self.set_SOC_state(SOC_state=self.standard_SOC_min)
        else:
            self.set_SOC_state(SOC_state=self.SOC_start)
        self.state = BatteryState.ACTIVE
        if save_flag:
            self.initialize_history()

    def set_mode(self, mode):
        """
        Sets the microgrid mode

        Parameters
        ----------
        mode : str
            The microgrid mode

        Returns
        ----------
        None

        """
        self.mode = mode

    def start_survival_time(self):
        """
        Starts the timer for how long the battery should focus on supporting own load
        Only for when a microgrid is added and follows a survival mode

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.remaining_survival_time = self.survival_time

    def set_SOC_state(self, SOC_state: float):
        """
        Sets the SOC state

        Parameters
        ----------
        SOC_state : float
            SOC value, between SOC_min and SOC_max

        Returns
        ----------
        None

        """
        if SOC_state < self.SOC_min or SOC_state > self.SOC_max:
            raise Exception("Not a valid SOC state")
        self.E_battery = SOC_state * self.E_max
        self.update_SOC()

    def draw_SOC_state(self):
        """
        Draws the SOC state based on a uniform distribution

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.E_battery = self.ps_random.uniform(
            low=self.E_min,
            high=self.E_max,
        )
        self.update_SOC()

    def update(self, p: float, q: float, fail_duration: Time, dt: Time):
        """
        Updates the battery status for the current time step

        Parameters
        ----------
        p : float
            Active power balance of the parent power system
        q : float
            Reactive power balance of the parent power system
        fail_duration : Time
            The duration of the current failure
        dt : Time
            The current time step

        Returns
        ----------
        p : float
            Remaining active power balance of the parent power system
        q : float
            Remaining reactive power balance of the parent power system
        """
        if (
            self.mode in [MicrogridMode.SURVIVAL, MicrogridMode.FULL_SUPPORT]
            and fail_duration == dt
        ):
            # Special handling for microgrid modes
            # Failure occurred in current time step
            self.draw_SOC_state()
        p, q = self.update_bus_load_and_prod(p, q, dt)
        return p, q


if __name__ == "__main__":
    pass



from relsad.network.components import ICTLine, ICTNode


def is_connected(
    node_1: ICTNode,
    node_2: ICTNode,
    network,
):
    """
    Function that checks if node 1 and node 2 are connected using
    a depth first search

    Parameters
    ----------
    node_1 : ICTNode
        ICT node 1
    node_2 : ICTNode
        ICT node 2
    network : ICTNetwork
        ICT network

    Returns
    -------
    connected : bool
        Boolean variable stating whether the ICT nodes are connected

    """
    connected = False

    visited_nodes = {node: False for node in network.nodes}

    stack = [node_1]

    while len(stack) > 0:
        node = stack.pop()
        if node == node_2:
            connected = True
        if visited_nodes[node] is False:
            visited_nodes[node] = True
        for neighbor_node in node.get_neighbor_nodes():
            if visited_nodes[neighbor_node] is False:
                stack.append(neighbor_node)

    return connected



import os
import sys

from relsad.examples.load.load_and_gen_data import (
    load_data,
    pv_power,
    weather_generation_data,
    wind_power,
)
from relsad.load.bus import CostFunction
from relsad.network.systems import PowerSystem
from relsad.Time import Time, TimeStamp, TimeUnit


def set_network_load_and_prod(
    power_system: PowerSystem,
    data_dir=os.path.join(
        os.pardir,
        os.pardir,
        "data",
    ),
):

    # Fetching bus-objects
    B2 = power_system.get_comp("B2")
    B3 = power_system.get_comp("B3")
    B4 = power_system.get_comp("B4")
    B5 = power_system.get_comp("B5")
    B6 = power_system.get_comp("B6")
    B7 = power_system.get_comp("B7")
    B8 = power_system.get_comp("B8")
    B9 = power_system.get_comp("B9")
    B10 = power_system.get_comp("B10")
    B11 = power_system.get_comp("B11")
    B12 = power_system.get_comp("B12")
    B13 = power_system.get_comp("B13")
    B14 = power_system.get_comp("B14")
    B15 = power_system.get_comp("B15")
    B16 = power_system.get_comp("B16")

    BM2 = power_system.get_comp("BM2")
    BM3 = power_system.get_comp("BM3")
    BM4 = power_system.get_comp("BM4")

    # Fetching production objects
    P1 = BM3.get_production()
    P2 = BM4.get_production()

    temp_profiles, wind_profiles, solar_profiles = weather_generation_data(
        path=os.path.join(data_dir, "weather_data_rygge.csv"),
    )

    wind = wind_power(wind_profiles)
    PV = pv_power(temp_profiles, solar_profiles)

    (
        load_household,
        load_farm,
        load_microgrid,
        load_industry2,
        load_trade,
        load_office,
    ) = load_data(
        temp_profiles,
        path=os.path.join(data_dir, "load_profiles_fasit.csv"),
    )

    farm = CostFunction(
        A=21.4 - 17.5,
        B=17.5,
    )

    industry = CostFunction(
        A=132.6 - 92.5,
        B=92.5,
    )

    trade = CostFunction(
        A=220.3 - 102.4,
        B=102.4,
    )

    public = CostFunction(
        A=194.5 - 31.4,
        B=31.4,
    )

    household = CostFunction(
        A=8.8,
        B=14.7,
    )

    for bus in [B2, B4, B13, B14, B16]:
        bus.add_load_data(pload_data=load_household, cost_function=household)

    for bus in [B6, B7, B11, B15, BM2]:
        bus.add_load_data(pload_data=load_farm, cost_function=farm)

    for bus in [B5, B10]:
        bus.add_load_data(pload_data=load_industry2, cost_function=industry)

    for bus in [B3, B8]:
        bus.add_load_data(pload_data=load_office, cost_function=public)

    for bus in [B9, B12]:
        bus.add_load_data(pload_data=load_trade, cost_function=trade)

    P1.add_prod_data(
        pprod_data=PV,
    )
    P2.add_prod_data(
        pprod_data=wind,
    )
    return power_system



from .system import (
    SAIFI,
    SAIDI,
    CAIDI,
    ASUI,
    ASAI,
    ENS,
)

from .ev import (
    EV_Index,
    EV_Interruption,
    EV_Duration,
)

__all__ = []
for v in dir():
    if not v.startswith("__") and v != "relsad":
        __all__.append(v)



import os

from relsad.examples.load.load_and_gen_data import (
    load_data,
    pv_power,
    weather_generation_data,
    wind_power,
)
from relsad.load.bus import CostFunction
from relsad.network.systems import PowerSystem
from relsad.Time import Time, TimeStamp, TimeUnit


def set_network_load_and_prod(
    power_system: PowerSystem,
    include_microgrid: bool = False,
    data_dir=os.path.join(
        os.pardir,
        os.pardir,
        "data",
    ),
):

    # Fetching bus-objects
    B2 = power_system.get_comp("B2")
    B3 = power_system.get_comp("B3")
    B4 = power_system.get_comp("B4")
    B5 = power_system.get_comp("B5")
    B6 = power_system.get_comp("B6")
    B7 = power_system.get_comp("B7")
    B8 = power_system.get_comp("B8")
    B9 = power_system.get_comp("B9")
    B10 = power_system.get_comp("B10")
    B11 = power_system.get_comp("B11")
    B12 = power_system.get_comp("B12")
    B13 = power_system.get_comp("B13")
    B14 = power_system.get_comp("B14")
    B15 = power_system.get_comp("B15")
    B16 = power_system.get_comp("B16")
    B17 = power_system.get_comp("B17")
    B18 = power_system.get_comp("B18")
    B19 = power_system.get_comp("B19")
    B20 = power_system.get_comp("B20")
    B21 = power_system.get_comp("B21")
    B22 = power_system.get_comp("B22")
    B23 = power_system.get_comp("B23")
    B24 = power_system.get_comp("B24")
    B25 = power_system.get_comp("B25")
    B26 = power_system.get_comp("B26")
    B27 = power_system.get_comp("B27")
    B28 = power_system.get_comp("B28")
    B29 = power_system.get_comp("B29")
    B30 = power_system.get_comp("B30")
    B31 = power_system.get_comp("B31")
    B32 = power_system.get_comp("B32")
    B33 = power_system.get_comp("B33")

    # Fetching battery and production objects

    temp_profiles, wind_profiles, solar_profiles = weather_generation_data(
        path=os.path.join(data_dir, "weather_data_rygge.csv"),
    )

    wind = wind_power(wind_profiles)
    PV = pv_power(temp_profiles, solar_profiles)

    (
        load_house,
        load_farm,
        load_microgrid,
        load_industry2,
        load_trade,
        load_office,
    ) = load_data(
        temp_profiles,
        path=os.path.join(data_dir, "load_profiles_fasit.csv"),
    )

    farm = CostFunction(
        A=21.4 - 17.5,
        B=17.5,
    )

    microgrid = CostFunction(
        A=(21.4 - 17.5) * 1000,
        B=17.5 * 1000,
    )

    industry = CostFunction(
        A=132.6 - 92.5,
        B=92.5,
    )

    trade = CostFunction(
        A=220.3 - 102.4,
        B=102.4,
    )

    household = CostFunction(
        A=8.8,
        B=14.7,
    )

    for bus in [B3, B9, B10, B11, B15, B16, B17, B19, B20, B23, B26, B27]:
        bus.add_load_data(
            pload_data=load_house,
            cost_function=household,
        )
    for bus in [B2, B29]:
        bus.add_load_data(
            pload_data=load_trade,
            cost_function=trade,
        )
    for bus in [B4, B14, B31]:
        bus.add_load_data(
            pload_data=load_office,
            cost_function=trade,
        )
    for bus in [B5, B6, B12, B13, B18, B21, B22, B28, B33]:
        bus.add_load_data(
            pload_data=load_farm,
            cost_function=farm,
        )
    for bus in [B7, B8, B24, B25, B30, B32]:
        bus.add_load_data(
            pload_data=load_industry2,
            cost_function=industry,
        )

    if include_microgrid:

        BM2 = power_system.get_comp("BM2")
        BM3 = power_system.get_comp("BM3")
        BM4 = power_system.get_comp("BM4")

        P1 = BM3.get_production()
        P2 = BM4.get_production()

        P1.add_prod_data(
            pprod_data=wind,
        )
        P2.add_prod_data(
            pprod_data=2 * PV,
        )

        BM2.add_load_data(
            pload_data=load_microgrid,
            cost_function=microgrid,
        )
    return power_system



from relsad.network.components import (
    Bus,
    CircuitBreaker,
    Disconnector,
    Line,
    ManualMainController,
)
from relsad.network.systems import Distribution, PowerSystem, Transmission
from relsad.StatDist import NormalParameters, StatDist, StatDistType
from relsad.Table import Table
from relsad.Time import Time, TimeUnit
from relsad.visualization.plotting import plot_topology


def initialize_network():

    C1 = ManualMainController(
        name="C1",
        sectioning_time=Time(1, TimeUnit.HOUR),
    )
    ps = PowerSystem(C1)

    line_stat_dist = StatDist(
        stat_dist_type=StatDistType.TRUNCNORMAL,
        parameters=NormalParameters(
            loc=1.25,
            scale=1,
            min_val=0.5,
            max_val=2,
        ),
    )
    fail_rate_trafo = 0.007  # 0.008
    fail_rate_line = 0.07  # 0.08

    B1 = Bus(
        name="B1",
        coordinate=[0, 0],
        fail_rate_per_year=0,
    )
    B2 = Bus(
        name="B2",
        coordinate=[0, -1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B3 = Bus(
        name="B3",
        coordinate=[0, -2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B4 = Bus(
        name="B4",
        coordinate=[0, -3],
        fail_rate_per_year=fail_rate_trafo,
    )
    B5 = Bus(
        name="B5",
        coordinate=[0, -4],
        fail_rate_per_year=fail_rate_trafo,
    )
    B6 = Bus(
        name="B6",
        coordinate=[0, -5],
        fail_rate_per_year=fail_rate_trafo,
    )
    B7 = Bus(
        name="B7",
        coordinate=[0, -6],
        fail_rate_per_year=fail_rate_trafo,
    )
    B8 = Bus(
        name="B8",
        coordinate=[0, -7],
        fail_rate_per_year=fail_rate_trafo,
    )
    B9 = Bus(
        name="B9",
        coordinate=[0, -8],
        fail_rate_per_year=fail_rate_trafo,
    )
    B10 = Bus(
        name="B10",
        coordinate=[0, -9],
        fail_rate_per_year=fail_rate_trafo,
    )
    B11 = Bus(
        name="B11",
        coordinate=[0, -10],
        fail_rate_per_year=fail_rate_trafo,
    )
    B12 = Bus(
        name="B12",
        coordinate=[0, -11],
        fail_rate_per_year=fail_rate_trafo,
    )
    B13 = Bus(
        name="B13",
        coordinate=[0, -12],
        fail_rate_per_year=fail_rate_trafo,
    )
    B14 = Bus(
        name="B14",
        coordinate=[0, -13],
        fail_rate_per_year=fail_rate_trafo,
    )
    B15 = Bus(
        name="B15",
        coordinate=[0, -14],
        fail_rate_per_year=fail_rate_trafo,
    )
    B16 = Bus(
        name="B16",
        coordinate=[0, -15],
        fail_rate_per_year=fail_rate_trafo,
    )
    B17 = Bus(
        name="B17",
        coordinate=[0, -16],
        fail_rate_per_year=fail_rate_trafo,
    )
    B18 = Bus(
        name="B18",
        coordinate=[0, -17],
        fail_rate_per_year=fail_rate_trafo,
    )
    B19 = Bus(
        name="B19",
        coordinate=[-1, -2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B20 = Bus(
        name="B20",
        coordinate=[-1, -3],
        fail_rate_per_year=fail_rate_trafo,
    )
    B21 = Bus(
        name="B21",
        coordinate=[-1, -4],
        fail_rate_per_year=fail_rate_trafo,
    )
    B22 = Bus(
        name="B22",
        coordinate=[-1, -5],
        fail_rate_per_year=fail_rate_trafo,
    )
    B23 = Bus(
        name="B23",
        coordinate=[1, -3],
        fail_rate_per_year=fail_rate_trafo,
    )
    B24 = Bus(
        name="B24",
        coordinate=[1, -4],
        fail_rate_per_year=fail_rate_trafo,
    )
    B25 = Bus(
        name="B25",
        coordinate=[1, -5],
        fail_rate_per_year=fail_rate_trafo,
    )
    B26 = Bus(
        name="B26",
        coordinate=[-1, -6],
        fail_rate_per_year=fail_rate_trafo,
    )
    B27 = Bus(
        name="B27",
        coordinate=[-1, -7],
        fail_rate_per_year=fail_rate_trafo,
    )
    B28 = Bus(
        name="B28",
        coordinate=[-1, -8],
        fail_rate_per_year=fail_rate_trafo,
    )
    B29 = Bus(
        name="B29",
        coordinate=[-1, -9],
        fail_rate_per_year=fail_rate_trafo,
    )
    B30 = Bus(
        name="B30",
        coordinate=[-1, -10],
        fail_rate_per_year=fail_rate_trafo,
    )
    B31 = Bus(
        name="B31",
        coordinate=[-1, -11],
        fail_rate_per_year=fail_rate_trafo,
    )
    B32 = Bus(
        name="B32",
        coordinate=[-1, -12],
        fail_rate_per_year=fail_rate_trafo,
    )
    B33 = Bus(
        name="B33",
        coordinate=[-1, -13],
        fail_rate_per_year=fail_rate_trafo,
    )

    L1 = Line(
        "L1",
        B1,
        B2,
        0.057526629463617,
        0.029324854498807,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L2 = Line(
        "L2",
        B2,
        B3,
        0.307599005700253,
        0.156669594992563,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L3 = Line(
        "L3",
        B3,
        B4,
        0.228359505246029,
        0.11630112507612,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L4 = Line(
        "L4",
        B4,
        B5,
        0.237780894670114,
        0.121105409749329,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L5 = Line(
        "L5",
        B5,
        B6,
        0.511001187968574,
        0.441120683630991,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L6 = Line(
        "L6",
        B6,
        B7,
        0.116800271535674,
        0.386089786465145,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L7 = Line(
        "L7",
        B7,
        B8,
        1.06779906360124,
        0.770619740244183,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L8 = Line(
        "L8",
        B8,
        B9,
        0.642651066675984,
        0.4617104750876,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L9 = Line(
        "L9",
        B9,
        B10,
        0.651386129718182,
        0.4617104750876,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L10 = Line(
        "L10",
        B10,
        B11,
        0.122665242435435,
        0.040555649838776,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L11 = Line(
        "L11",
        B11,
        B12,
        0.233600543071348,
        0.077242914616007,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L12 = Line(
        "L12",
        B12,
        B13,
        0.915933753281888,
        0.720642700981322,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L13 = Line(
        "L13",
        B13,
        B14,
        0.337922153118168,
        0.444801888770203,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L14 = Line(
        "L14",
        B14,
        B15,
        0.368744446995637,
        0.328188797156862,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L15 = Line(
        "L15",
        B15,
        B16,
        0.465641253456589,
        0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L16 = Line(
        "L16",
        B16,
        B17,
        0.804249732956644,
        1.07378882111589,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L17 = Line(
        "L17",
        B17,
        B18,
        0.456719010492059,
        0.358137584730111,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L18 = Line(
        "L18",
        B2,
        B19,
        0.102325024208603,
        0.097645526150283,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L19 = Line(
        "L19",
        B19,
        B20,
        0.938520130576714,
        0.84567888909964,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L20 = Line(
        "L20",
        B20,
        B21,
        0.255500593984287,
        0.298489582813389,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L21 = Line(
        "L21",
        B21,
        B22,
        0.442306156472432,
        0.584812470675146,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L22 = Line(
        "L22",
        B3,
        B23,
        0.281518603188548,
        0.192358566850685,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L23 = Line(
        "L23",
        B23,
        B24,
        0.560291900849547,
        0.442430943087321,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L24 = Line(
        "L24",
        B24,
        B25,
        0.559044034700662,
        0.437439478491779,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L25 = Line(
        "L25",
        B6,
        B26,
        0.126658414111869,
        0.064514679897376,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L26 = Line(
        "L26",
        B26,
        B27,
        0.177321779756616,
        0.090283115871859,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L27 = Line(
        "L27",
        B27,
        B28,
        0.660745125834823,
        0.582566311607152,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L28 = Line(
        "L28",
        B28,
        B29,
        0.501766978466822,
        0.437127511954558,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L29 = Line(
        "L29",
        B29,
        B30,
        0.316646035279672,
        0.161286699743439,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L30 = Line(
        "L30",
        B30,
        B31,
        0.60796038773697,
        0.600847550688323,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L31 = Line(
        "L31",
        B31,
        B32,
        0.193731219614459,
        0.225801379640814,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L32 = Line(
        "L32",
        B32,
        B33,
        0.212761178384962,
        0.330809316069521,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )

    CircuitBreaker("E1", L1)

    Disconnector("L1a", L1, B1)
    Disconnector("L1b", L1, B2)
    Disconnector("L2a", L2, B2)
    Disconnector("L2b", L2, B3)
    Disconnector("L3a", L3, B3)
    Disconnector("L3b", L3, B4)
    Disconnector("L4a", L4, B4)
    Disconnector("L4b", L4, B5)
    Disconnector("L5a", L5, B5)
    Disconnector("L5b", L5, B6)
    Disconnector("L6a", L6, B6)
    Disconnector("L6b", L6, B7)
    Disconnector("L7a", L7, B7)
    Disconnector("L7b", L7, B8)
    Disconnector("L8a", L8, B8)
    Disconnector("L8b", L8, B9)
    Disconnector("L9a", L9, B9)
    Disconnector("L9b", L9, B10)
    Disconnector("L10a", L10, B10)
    Disconnector("L10b", L10, B11)
    Disconnector("L11a", L11, B11)
    Disconnector("L11b", L11, B12)
    Disconnector("L12a", L12, B12)
    Disconnector("L12b", L12, B13)
    Disconnector("L13a", L13, B13)
    Disconnector("L13b", L13, B14)
    Disconnector("L14a", L14, B14)
    Disconnector("L14b", L14, B15)
    Disconnector("L15a", L15, B15)
    Disconnector("L15b", L15, B16)
    Disconnector("L16a", L16, B16)
    Disconnector("L16b", L16, B17)
    Disconnector("L17a", L17, B17)
    Disconnector("L17b", L17, B18)

    Disconnector("L18a", L18, B2)
    Disconnector("L18b", L18, B19)
    Disconnector("L19a", L19, B19)
    Disconnector("L19b", L19, B20)
    Disconnector("L20a", L20, B20)
    Disconnector("L20b", L20, B21)
    Disconnector("L21a", L21, B21)
    Disconnector("L21b", L21, B22)

    Disconnector("L22a", L22, B3)
    Disconnector("L22b", L22, B23)
    Disconnector("L23a", L23, B23)
    Disconnector("L23b", L23, B24)
    Disconnector("L24a", L24, B24)
    Disconnector("L24b", L24, B25)

    Disconnector("L25a", L25, B6)
    Disconnector("L25b", L25, B26)
    Disconnector("L26a", L26, B26)
    Disconnector("L26b", L26, B27)
    Disconnector("L27a", L27, B27)
    Disconnector("L27b", L27, B28)
    Disconnector("L28a", L28, B28)
    Disconnector("L28b", L28, B29)
    Disconnector("L29a", L29, B29)
    Disconnector("L29b", L29, B30)
    Disconnector("L30a", L30, B30)
    Disconnector("L30b", L30, B31)
    Disconnector("L31a", L31, B31)
    Disconnector("L31b", L31, B32)
    Disconnector("L32a", L32, B32)
    Disconnector("L32b", L32, B33)

    tn = Transmission(ps, B1)

    dn = Distribution(tn, L1)

    dn.add_buses(
        [
            B2,
            B3,
            B4,
            B5,
            B6,
            B7,
            B8,
            B9,
            B10,
            B11,
            B12,
            B13,
            B14,
            B15,
            B16,
            B17,
            B18,
            B19,
            B20,
            B21,
            B22,
            B23,
            B24,
            B25,
            B26,
            B27,
            B28,
            B29,
            B30,
            B31,
            B32,
            B33,
        ]
    )

    dn.add_lines(
        [
            L2,
            L3,
            L4,
            L5,
            L6,
            L7,
            L8,
            L9,
            L10,
            L11,
            L12,
            L13,
            L14,
            L15,
            L16,
            L17,
            L18,
            L19,
            L20,
            L21,
            L22,
            L23,
            L24,
            L25,
            L26,
            L27,
            L28,
            L29,
            L30,
            L31,
            L32,
        ]
    )

    return ps


if __name__ == "__main__":
    import os

    ps = initialize_network()
    fig = plot_topology(ps.buses, ps.lines)

    fig.savefig(
        os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "RBTS6.pdf",
        )
    )



from relsad.examples.tutorial.system_components import *

from relsad.network.components import (
    Production,
    Battery,
    EVPark,
)
from relsad.Table import Table

import numpy as np

# A generation unit:

P1 = Production(
    name="P1",
    bus=B3,
)

# A battery:

Bat1 = Battery(
    name="B1",
    bus=B6,
)

# An EV park

num_ev_table = Table(
    x=np.arange(0, 24),  # Hour of the day
    y=np.ones(24) * 10,  # Number of EVs
)

EVPark(
    name="EV1",
    bus=B5,
    num_ev_dist=num_ev_table,
    v2g_flag=True,
)

generation_profile = np.ones(365 * 24) * 0.02  # MW
P1.add_prod_data(
    pprod_data=generation_profile,
)



from collections import namedtuple
from enum import Enum

import matplotlib.pyplot as plt
import numpy as np
import scipy.special as sps
from scipy import stats

from relsad.utils import get_random_instance


class StatDistType(Enum):
    """
    Statistical distribution type

    Attributes
    ----------
    UNIFORM_FLOAT : int
        Uniform distribution with floating point numbers.
        Parameters from UniformParameters:
        (min_val, max_val)
    UNIFORM_INT : int
        Uniform distribution with integers.
        Parameters from UniformParameters:
        (min_val, max_val)
    TRUNCNORMAL : int
        Truncated normal distribution.
        Parameters from NormalParameters:
        (loc, scale, min_val, max_val)
    GAMMA : int
        Gamma distribution
        Parameters from GammaParameters:
        (shape, scale)
    """

    UNIFORM_FLOAT = 1
    UNIFORM_INT = 2
    TRUNCNORMAL = 3
    GAMMA = 4


UniformParameters = namedtuple(
    "UniformParameters",
    ["min_val", "max_val"],
)
NormalParameters = namedtuple(
    "NormalParameters",
    ["loc", "scale", "min_val", "max_val"],
)
GammaParameters = namedtuple(
    "GammaParameters",
    ["shape", "scale"],
)


class StatDist:

    """
    Utility class for statistical distributions in relsad

    ...

    Attributes
    ----------
    stat_dist_type : StatDistType
        Type of statistical distribution
    parameters : namedtuple
        Statistical distribution parameters


    Methods
    ----------
    draw(random_instance, size)
        Returns array of drawn instances of the statistical distribution
        of given size
    get_pdf(x)
        Returns the probability distribution function of
        the statistical distribution
    histplot(ax, n_points, n_bins)
        Plots the statistical distribution in a histogram
    plot(ax, x, color)

    """

    def __init__(
        self,
        stat_dist_type: StatDistType,
        parameters: namedtuple,
    ):
        self.stat_dist_type = stat_dist_type
        self.parameters = parameters

    def draw(self, random_instance, size: int = 1):
        """
        Returns array of drawn instances of the statistical distribution
        of given size

        Parameters
        ----------
        random_instance : np.random.Generator
            Instance of a random generator
        size : int
            Size of drawn values

        Returns
        ----------
        drawn_values : np.ndarray
            Array of drawn instances of the statistical distribution
            of given size

        """
        if random_instance is None:
            random_instance = np.random.default_rng()
        drawn_values = None
        if self.stat_dist_type == StatDistType.UNIFORM_FLOAT:
            drawn_values = random_instance.uniform(
                low=self.parameters.min_val,
                high=self.parameters.max_val,
                size=size,
            )
        elif self.stat_dist_type == StatDistType.UNIFORM_INT:
            drawn_values = random_instance.integers(
                low=self.parameters.min_val,
                high=self.parameters.max_val,
                size=size,
            )
        elif self.stat_dist_type == StatDistType.TRUNCNORMAL:
            drawn_values = stats.truncnorm.rvs(
                (self.parameters.min_val - self.parameters.loc)
                / self.parameters.scale,
                (self.parameters.max_val - self.parameters.loc)
                / self.parameters.scale,
                loc=self.parameters.loc,
                scale=self.parameters.scale,
                size=size,
                random_state=random_instance,
            )
        elif self.stat_dist_type == StatDistType.GAMMA:
            drawn_values = random_instance.gamma(
                shape=self.parameters.shape,
                scale=self.parameters.scale,
                size=size,
            )
        return drawn_values

    def get_pdf(
        self,
        x,
    ):
        """
        Returns the probability distribution function of
        the statistical distribution

        Parameters
        ----------
        x : np.ndarray
            Array of possible distribution values

        Returns
        ----------
        None

        """
        if self.stat_dist_type == StatDistType.UNIFORM_FLOAT:
            pass
        elif self.stat_dist_type == StatDistType.UNIFORM_INT:
            pass
        elif self.stat_dist_type == StatDistType.TRUNCNORMAL:
            return stats.truncnorm.pdf(
                x,
                (self.parameters.min_val - self.parameters.loc)
                / self.parameters.scale,
                (self.parameters.max_val - self.parameters.loc)
                / self.parameters.scale,
                loc=self.parameters.loc,
                scale=self.parameters.scale,
            )
        elif self.stat_dist_type == StatDistType.GAMMA:
            return x ** (self.parameters.shape - 1) * (
                np.exp(-x / self.parameters.scale)
                / (
                    sps.gamma(self.parameters.shape)
                    * self.parameters.scale**self.parameters.shape
                )
            )

    def histplot(
        self,
        ax,
        n_points: int = 5000,
        n_bins: int = 50,
    ):
        """
        Plots the statistical distribution in a normalized histogram

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Plot axis
        n_points : int
            Number of distribution points
        n_bins : int
            Number of bins in histogram

        Returns
        ----------
        None

        """
        random_instance = get_random_instance()
        dist = self.draw(
            random_instance,
            size=n_points,
        )
        ax.hist(dist, bins=n_bins, density=True)

    def plot(
        self,
        ax,
        x,
        color: str = "b",
        label: str = None,
    ):
        """
        Plots the statistical distribution in the provided axis

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Plot axis
        x : np.ndarray
            Array of possible distribution values
        color : str

        Returns
        ----------
        None

        """
        if self.stat_dist_type == StatDistType.UNIFORM_FLOAT:
            pass
        elif self.stat_dist_type == StatDistType.UNIFORM_INT:
            pass
        elif self.stat_dist_type == StatDistType.TRUNCNORMAL:
            ax.plot(
                x,
                self.get_pdf(x),
                color=color,
                label=label,
            )
        elif self.stat_dist_type == StatDistType.GAMMA:
            ax.plot(
                x,
                self.get_pdf(x),
                color=color,
                label=label,
            )


if __name__ == "__main__":
    trunc_normal_loc = 2
    trunc_normal_scale = 1
    min_val, max_val = 0, 20
    trunc_normal = StatDist(
        stat_dist_type=StatDistType.TRUNCNORMAL,
        parameters=NormalParameters(
            loc=trunc_normal_loc,
            scale=trunc_normal_scale,
            min_val=min_val,
            max_val=max_val,
        ),
    )

    gamma_1_shape = 1
    gamma_1_scale = 2
    gamma_1 = StatDist(
        stat_dist_type=StatDistType.GAMMA,
        parameters=GammaParameters(
            shape=gamma_1_shape,
            scale=gamma_1_scale,
        ),
    )

    gamma_2_shape = 7.5
    gamma_2_scale = 1
    gamma_2 = StatDist(
        stat_dist_type=StatDistType.GAMMA,
        parameters=GammaParameters(
            shape=gamma_2_shape,
            scale=gamma_2_scale,
        ),
    )
    fig, ax = plt.subplots()
    x = np.linspace(min_val, max_val, 1000)
    trunc_normal.plot(ax=ax, x=x)
    gamma_1.plot(ax=ax, x=x)
    gamma_1.histplot(ax=ax)
    gamma_2.plot(ax=ax, x=x)
    gamma_2.histplot(ax=ax)
    fig.savefig(fname="stat_dist_plot.pdf")
    plt.show()



from relsad.network.components import (
    Battery,
    Bus,
    CircuitBreaker,
    Disconnector,
    IntelligentSwitch,
    Line,
    ManualMainController,
    Production,
    Sensor,
)
from relsad.network.systems import Distribution, PowerSystem, Transmission
from relsad.StatDist import NormalParameters, StatDist, StatDistType
from relsad.Table import Table
from relsad.Time import Time, TimeUnit
from relsad.visualization.plotting import plot_topology


def initialize_network():
    C1 = ManualMainController(
        name="C1", sectioning_time=Time(1, TimeUnit.HOUR)
    )

    ps = PowerSystem(C1)
    fail_rate_trafo = 0.0150
    fail_rate_line = 0.0650

    line_stat_dist = StatDist(
        stat_dist_type=StatDistType.TRUNCNORMAL,
        parameters=NormalParameters(
            loc=1.25,
            scale=1,
            min_val=0.5,
            max_val=2,
        ),
    )

    rho = 1.72e-8
    a = 64.52e-6
    l1 = 0.6  # km
    l2 = 0.75  # km
    l3 = 0.8  # km

    r1 = (rho * l1) / a * 1e3
    r2 = (rho * l2) / a * 1e3
    r3 = (rho * l3) / a * 1e3

    B0 = Bus(
        "B0",
        n_customers=0,
        coordinate=[0, 0],
        fail_rate_per_year=0,
    )

    BF11 = Bus(
        "BF11",
        n_customers=0,
        coordinate=[-4, -1],
        fail_rate_per_year=0,
    )
    B1 = Bus(
        "B1",
        n_customers=210,
        coordinate=[-5, -2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B2 = Bus(
        "B2",
        n_customers=210,
        coordinate=[-5, -3],
        fail_rate_per_year=fail_rate_trafo,
    )
    BF12 = Bus(
        "BF12",
        n_customers=0,
        coordinate=[-4, -4],
        fail_rate_per_year=0,
    )
    B3 = Bus(
        "B3",
        n_customers=210,
        coordinate=[-5, -5],
        fail_rate_per_year=fail_rate_trafo,
    )
    B4 = Bus(
        "B4",
        n_customers=1,
        coordinate=[-5, -6],
        fail_rate_per_year=fail_rate_trafo,
    )
    BF13 = Bus(
        "BF13",
        n_customers=0,
        coordinate=[-4, -7],
        fail_rate_per_year=0,
    )
    B5 = Bus(
        "B5",
        n_customers=1,
        coordinate=[-5, -8],
        fail_rate_per_year=fail_rate_trafo,
    )
    B6 = Bus(
        "B6",
        n_customers=10,
        coordinate=[-5, -9],
        fail_rate_per_year=fail_rate_trafo,
    )
    BF14 = Bus(
        "BF14",
        n_customers=0,
        coordinate=[-4, -10],
        fail_rate_per_year=0,
    )
    B7 = Bus(
        "B7",
        n_customers=10,
        coordinate=[-5, -11],
        fail_rate_per_year=fail_rate_trafo,
    )

    B8 = Bus(
        "B8",
        n_customers=1,
        coordinate=[-3, -2],
        fail_rate_per_year=0,
    )
    BF21 = Bus(
        "BF21",
        n_customers=0,
        coordinate=[-2, -1],
        fail_rate_per_year=0,
    )
    BF22 = Bus(
        "BF22",
        n_customers=0,
        coordinate=[-2, -3],
        fail_rate_per_year=0,
    )
    B9 = Bus(
        "B9",
        n_customers=1,
        coordinate=[-3, -4],
        fail_rate_per_year=fail_rate_trafo,
    )

    BF31 = Bus(
        "BF31",
        n_customers=0,
        coordinate=[2, -1],
        fail_rate_per_year=0,
    )
    B10 = Bus(
        "B10",
        n_customers=210,
        coordinate=[3, -2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B11 = Bus(
        "B11",
        n_customers=210,
        coordinate=[3, -3],
        fail_rate_per_year=fail_rate_trafo,
    )
    BF32 = Bus(
        "BF32",
        n_customers=0,
        coordinate=[2, -4],
        fail_rate_per_year=0,
    )
    B12 = Bus(
        "B12",
        n_customers=200,
        coordinate=[3, -5],
        fail_rate_per_year=fail_rate_trafo,
    )
    B13 = Bus(
        "B13",
        n_customers=1,
        coordinate=[3, -6],
        fail_rate_per_year=fail_rate_trafo,
    )
    BF33 = Bus(
        "BF33",
        n_customers=0,
        coordinate=[2, -7],
        fail_rate_per_year=0,
    )
    B14 = Bus(
        "B14",
        n_customers=1,
        coordinate=[3, -8],
        fail_rate_per_year=fail_rate_trafo,
    )
    BF34 = Bus(
        "BF34",
        n_customers=0,
        coordinate=[2, -9],
        fail_rate_per_year=0,
    )
    B15 = Bus(
        "B15",
        n_customers=10,
        coordinate=[3, -10],
        fail_rate_per_year=0,
    )

    BF41 = Bus(
        "BF41",
        n_customers=0,
        coordinate=[4, -1],
        fail_rate_per_year=0,
    )
    B16 = Bus(
        "B16",
        n_customers=10,
        coordinate=[5, -2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B17 = Bus(
        "B17",
        n_customers=200,
        coordinate=[5, -3],
        fail_rate_per_year=fail_rate_trafo,
    )
    BF42 = Bus(
        "BF42",
        n_customers=0,
        coordinate=[4, -4],
        fail_rate_per_year=0,
    )
    B18 = Bus(
        "B18",
        n_customers=200,
        coordinate=[5, -5],
        fail_rate_per_year=fail_rate_trafo,
    )
    B19 = Bus(
        "B19",
        n_customers=200,
        coordinate=[5, -6],
        fail_rate_per_year=fail_rate_trafo,
    )
    BF43 = Bus(
        "BF43",
        n_customers=0,
        coordinate=[4, -7],
        fail_rate_per_year=0,
    )
    B20 = Bus(
        "B20",
        n_customers=1,
        coordinate=[5, -8],
        fail_rate_per_year=fail_rate_trafo,
    )
    B21 = Bus(
        "B21",
        n_customers=1,
        coordinate=[5, -9],
        fail_rate_per_year=fail_rate_trafo,
    )
    BF44 = Bus(
        "BF44",
        n_customers=0,
        coordinate=[4, -10],
        fail_rate_per_year=0,
    )
    B22 = Bus(
        "B22",
        n_customers=10,
        coordinate=[5, -11],
        fail_rate_per_year=fail_rate_trafo,
    )

    L1 = Line(
        "L1",
        B0,
        BF11,
        r=r2,
        x=0.156669594992563,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L2 = Line(
        "L2",
        BF11,
        B1,
        r=r1,
        x=0.156669594992563,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L3 = Line(
        "L3",
        BF11,
        B2,
        r=r3,
        x=0.11630112507612,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L4 = Line(
        "L4",
        BF11,
        BF12,
        r=r2,
        x=0.121105409749329,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L5 = Line(
        "L5",
        BF12,
        B3,
        r=r3,
        x=0.441120683630991,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L6 = Line(
        "L6",
        BF12,
        B4,
        r=r1,
        x=0.386089786465145,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L7 = Line(
        "L7",
        BF12,
        BF13,
        r=r2,
        x=0.770619740244183,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L8 = Line(
        "L8",
        BF13,
        B5,
        r=r3,
        x=0.4617104750876,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L9 = Line(
        "L9",
        BF13,
        B6,
        r=r2,
        x=0.4617104750876,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L10 = Line(
        "L10",
        BF13,
        BF14,
        r=r1,
        x=0.040555649838776,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L11 = Line(
        "L11",
        BF14,
        B7,
        r=r3,
        x=0.077242914616007,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    LB1 = Line(
        "LB1",
        BF14,
        BF22,
        r=r1,
        x=0.077242914616007,
        fail_rate_density_per_year=0,
        repair_time_dist=line_stat_dist,
    )

    L12 = Line(
        "L12",
        B0,
        BF21,
        r=r2,
        x=0.720642700981322,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )

    L13 = Line(
        "L13",
        BF21,
        B8,
        r=r3,
        x=0.444801888770203,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L14 = Line(
        "L14",
        BF21,
        BF22,
        r=r1,
        x=0.328188797156862,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L15 = Line(
        "L15",
        BF22,
        B9,
        r=r3,
        x=0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )

    L16 = Line(
        "L16",
        B0,
        BF31,
        r=r2,
        x=0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L17 = Line(
        "L17",
        BF31,
        B10,
        r=r1,
        x=0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L18 = Line(
        "L18",
        BF31,
        BF32,
        r=r3,
        x=0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L19 = Line(
        "L19",
        BF31,
        B11,
        r=r2,
        x=0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L20 = Line(
        "L20",
        BF32,
        B12,
        r=r3,
        x=0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L21 = Line(
        "L21",
        BF32,
        BF33,
        r=r1,
        x=0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L22 = Line(
        "L22",
        BF33,
        B13,
        r=r2,
        x=0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L23 = Line(
        "L23",
        BF33,
        B14,
        r=r3,
        x=0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L24 = Line(
        "L24",
        BF33,
        BF34,
        r=r2,
        x=0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L25 = Line(
        "L25",
        BF34,
        B15,
        r=r1,
        x=0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    LB2 = Line(
        "LB2",
        BF34,
        BF44,
        r=r1,
        x=0.340043525571273,
        fail_rate_density_per_year=0,
        repair_time_dist=line_stat_dist,
    )

    L26 = Line(
        "L26",
        B0,
        BF41,
        r=r3,
        x=0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L27 = Line(
        "L27",
        BF41,
        B16,
        r=r2,
        x=0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L28 = Line(
        "L28",
        BF41,
        B17,
        r=r1,
        x=0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L29 = Line(
        "L29",
        BF41,
        BF42,
        r=r2,
        x=0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L30 = Line(
        "L30",
        BF42,
        B18,
        r=r1,
        x=0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L31 = Line(
        "L31",
        BF42,
        B19,
        r=r3,
        x=0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L32 = Line(
        "L32",
        BF42,
        BF43,
        r=r2,
        x=0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L33 = Line(
        "L33",
        BF43,
        B20,
        r=r3,
        x=0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L34 = Line(
        "L34",
        BF43,
        BF44,
        r=r1,
        x=0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L35 = Line(
        "L35",
        BF44,
        B21,
        r=r2,
        x=0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L36 = Line(
        "L36",
        BF44,
        B22,
        r=r3,
        x=0.340043525571273,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )

    CircuitBreaker("E1", L1)
    CircuitBreaker("E2", L12)
    CircuitBreaker("E3", L16)
    CircuitBreaker("E4", L26)

    Disconnector("DL1a", L1, B0)
    Disconnector("DL1b", L1, BF11)
    Disconnector("DL4", L4, BF11)
    Disconnector("DL7", L7, BF12)
    Disconnector("DL10", L10, BF13)
    Disconnector("DLB1a", LB1, BF14)

    Disconnector("DL12a", L12, B0)
    Disconnector("DL12b", L12, BF21)
    Disconnector("DL14", L14, BF21)
    Disconnector("DLB1b", LB1, BF22)

    Disconnector("DL16a", L16, B0)
    Disconnector("DL16b", L16, BF31)
    Disconnector("DL18", L18, BF31)
    Disconnector("DL21", L21, BF32)
    Disconnector("DL24", L24, BF33)
    Disconnector("DLB2a", LB2, BF34)

    Disconnector("DL26a", L26, B0)
    Disconnector("DL26b", L26, BF41)
    Disconnector("DL29", L29, BF41)
    Disconnector("DL32", L32, BF42)
    Disconnector("DL34", L34, BF43)
    Disconnector("DLB2b", LB2, BF44)

    LB1.set_backup()
    LB2.set_backup()

    tn = Transmission(ps, B0)

    dn1 = Distribution(tn, L1)
    dn2 = Distribution(tn, L12)
    dn3 = Distribution(tn, L16)
    dn4 = Distribution(tn, L26)

    dn1.add_buses(
        [
            B0,
            BF11,
            B1,
            B2,
            BF12,
            B3,
            B4,
            BF13,
            B5,
            B6,
            BF14,
            B7,
        ]
    )
    dn2.add_buses(
        [
            BF21,
            B8,
            BF22,
            B9,
        ]
    )
    dn3.add_buses(
        [
            BF31,
            B10,
            BF32,
            B11,
            B12,
            BF33,
            B13,
            B14,
            BF34,
            B15,
        ]
    )

    dn4.add_buses(
        [
            BF41,
            B16,
            B17,
            BF42,
            B18,
            B19,
            BF43,
            B20,
            BF44,
            B21,
            B22,
        ]
    )

    dn1.add_lines(
        [
            L1,
            L2,
            L3,
            L4,
            L5,
            L6,
            L7,
            L8,
            L9,
            L10,
            L11,
            LB1,
        ]
    )
    dn2.add_lines(
        [
            L12,
            L13,
            L14,
            L15,
            LB1,
        ]
    )
    dn3.add_lines(
        [
            L16,
            L17,
            L18,
            L19,
            L20,
            L21,
            L22,
            L23,
            L24,
            L25,
            LB2,
        ]
    )
    dn4.add_lines(
        [
            L26,
            L27,
            L28,
            L29,
            L30,
            L31,
            L32,
            L33,
            L34,
            L35,
            L36,
            LB2,
        ]
    )

    return ps


if __name__ == "__main__":
    import os

    ps = initialize_network()
    fig = plot_topology(ps.buses, ps.lines)

    fig.savefig(
        os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "RBTS2.pdf",
        )
    )



import os
import time

from relsad.examples.IEEE16_modified.load_and_prod import (
    set_network_load_and_prod,
)
from relsad.examples.IEEE16_modified.network import initialize_network
from relsad.load.bus import CostFunction
from relsad.network.components import MicrogridMode
from relsad.simulation import Simulation
from relsad.StatDist import (
    GammaParameters,
    NormalParameters,
    StatDist,
    StatDistType,
)
from relsad.Time import Time, TimeStamp, TimeUnit


def run_simulation(
    fail_rate_trafo: float = 0.007,
    fail_rate_line: float = 0.7,
    microgrid_mode: MicrogridMode = MicrogridMode.SURVIVAL,
    line_repair_time_stat_dist: StatDist = StatDist(
        stat_dist_type=StatDistType.TRUNCNORMAL,
        parameters=NormalParameters(
            loc=1.25,
            scale=1,
            min_val=0.5,
            max_val=2,
        ),
    ),
    random_seed: int = 2837314,
    iterations: int = 2,
    start_time: TimeStamp = TimeStamp(
        year=2019,
        month=1,
        day=1,
        hour=0,
        minute=0,
        second=0,
    ),
    stop_time: TimeStamp = TimeStamp(
        year=2020,
        month=1,
        day=1,
        hour=0,
        minute=0,
        second=0,
    ),
    save_iterations: list = [1, 2],
    data_dir: str = os.path.join(
        os.pardir,
        "load",
        "data",
    ),
    save_flag: bool = True,
    save_dir: str = "results",
    n_procs: int = 1,
    debug: bool = True,
):

    start = time.time()

    ps = initialize_network(
        fail_rate_line=fail_rate_line,
        fail_rate_trafo=fail_rate_trafo,
        microgrid_mode=microgrid_mode,
        line_repair_time_stat_dist=line_repair_time_stat_dist,
    )

    ps = set_network_load_and_prod(
        power_system=ps,
        data_dir=data_dir,
    )

    sim = Simulation(ps, random_seed=random_seed)

    sim.run_monte_carlo(
        iterations=iterations,
        start_time=start_time,
        stop_time=stop_time,
        time_step=Time(5, TimeUnit.MINUTE),
        time_unit=TimeUnit.HOUR,
        save_iterations=save_iterations,
        save_dir=save_dir,
        n_procs=n_procs,
        debug=debug,
        save_flag=save_flag,
    )

    end = time.time()
    print("Time elapsed: {}".format(end - start))
